import sys
import timeit

import numpy as np
from scipy.integrate import cumulative_trapezoid
from scipy.interpolate import CubicSpline, interp1d
from scipy.optimize import least_squares
from tqdm import tqdm

from .bl import (
    calc_d1,
    calc_d2,
    calc_d3,
    calc_d99_1d,
    calc_dRC,
    calc_I2,
    calc_I3,
    calc_profile_edge_1d,
    calc_wake_parameter_1d,
)
from .gradient import gradient
from .grid_metric import get_metric_tensor_2d
from .utils import even_print, format_time_string

'''
========================================================================
Collection of routines that pre-compute useful quantities
========================================================================
'''

# ======================================================================

def _calc_gradients(self, acc=6, edge_stencil='full', **kwargs):
    '''
    Calculate spatial gradients of averaged quantities
    '''
    
    verbose      = kwargs.get('verbose',True)
    do_dtangnorm = kwargs.get('do_dtangnorm',False)
    
    if verbose: print('\n'+'ztmd.calc_gradients()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if verbose: even_print('acc','%i'%(acc,))
    if verbose: even_print('edge_stencil','%s'%(edge_stencil,))
    
    ## check
    if not hasattr(self,'rectilinear') and not hasattr(self,'curvilinear'):
        raise AssertionError('neither rectilinear nor curvilinear attr set')
    
    if hasattr(self,'rectilinear'):
        if self.rectilinear:
            if verbose: even_print('grid type','rectilinear')
    if hasattr(self,'curvilinear'):
        if self.curvilinear:
            if verbose: even_print('grid type','curvilinear')
    
    if (self.x.ndim==1) and (self.y.ndim==1):
        if hasattr(self,'rectilinear'):
            if not self.rectilinear:
                raise AssertionError
        if hasattr(self,'curvilinear'):
            if self.curvilinear:
                raise AssertionError
    elif (self.x.ndim==2) and (self.y.ndim==2):
        if hasattr(self,'rectilinear'):
            if self.rectilinear:
                raise AssertionError
        if hasattr(self,'curvilinear'):
            if not self.curvilinear:
                raise AssertionError
    else:
        raise ValueError
    
    # ===
    
    if self.curvilinear: ## get metric tensor 2D
        
        M = get_metric_tensor_2d(self.x, self.y, acc=acc, edge_stencil=edge_stencil, verbose=False)
        
        ddx_q1 = np.copy( M[:,:,0,0] ) ## ξ_x
        ddx_q2 = np.copy( M[:,:,1,0] ) ## η_x
        ddy_q1 = np.copy( M[:,:,0,1] ) ## ξ_y
        ddy_q2 = np.copy( M[:,:,1,1] ) ## η_y
        
        if verbose: even_print('ξ_x','%s'%str(ddx_q1.shape))
        if verbose: even_print('η_x','%s'%str(ddx_q2.shape))
        if verbose: even_print('ξ_y','%s'%str(ddy_q1.shape))
        if verbose: even_print('η_y','%s'%str(ddy_q2.shape))
        
        M = None; del M
        
        ## the 'computational' grid (unit Cartesian)
        #x_comp = np.arange(nx, dtype=np.float64)
        #y_comp = np.arange(ny, dtype=np.float64)
        x_comp = 1.
        y_comp = 1.
    
    print(72*'-')
    
    # ==================================================================
    # Get Cartesian gradients
    # ==================================================================
    
    if ('data/u' in self):
        
        u = np.copy( self['data/u'][()].T )
        
        if self.rectilinear:
            ddx_u = gradient(u, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_u = gradient(u, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_u_comp = gradient(u, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_u_comp = gradient(u, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_u      = ddx_u_comp*ddx_q1 + ddy_u_comp*ddx_q2
            ddy_u      = ddx_u_comp*ddy_q1 + ddy_u_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_u' in self): del self['data/ddx_u']
        self.create_dataset('data/ddx_u', data=ddx_u.T, chunks=None)
        
        if ('data/ddy_u' in self): del self['data/ddy_u']
        self.create_dataset('data/ddy_u', data=ddy_u.T, chunks=None)
        
        if verbose: even_print('∂x[u]','%s'%str(ddx_u.shape))
        if verbose: even_print('∂y[u]','%s'%str(ddy_u.shape))
    
    if ('data/v' in self):
        
        v = np.copy( self['data/v'][()].T )
        
        if self.rectilinear:
            ddx_v = gradient(v, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_v = gradient(v, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_v_comp = gradient(v, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_v_comp = gradient(v, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_v      = ddx_v_comp*ddx_q1 + ddy_v_comp*ddx_q2
            ddy_v      = ddx_v_comp*ddy_q1 + ddy_v_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_v' in self): del self['data/ddx_v']
        self.create_dataset('data/ddx_v', data=ddx_v.T, chunks=None)
        
        if ('data/ddy_v' in self): del self['data/ddy_v']
        self.create_dataset('data/ddy_v', data=ddy_v.T, chunks=None)
        
        if verbose: even_print('∂x[v]','%s'%str(ddx_v.shape))
        if verbose: even_print('∂y[v]','%s'%str(ddy_v.shape))
    
    if ('data/p' in self):
        
        p = np.copy( self['data/p'][()].T )
        
        if self.rectilinear:
            ddx_p = gradient(p, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_p = gradient(p, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_p_comp = gradient(p, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_p_comp = gradient(p, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_p      = ddx_p_comp*ddx_q1 + ddy_p_comp*ddx_q2
            ddy_p      = ddx_p_comp*ddy_q1 + ddy_p_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_p' in self): del self['data/ddx_p']
        self.create_dataset('data/ddx_p', data=ddx_p.T, chunks=None)
        
        if ('data/ddy_p' in self): del self['data/ddy_p']
        self.create_dataset('data/ddy_p', data=ddy_p.T, chunks=None)
        
        if verbose: even_print('∂x[p]','%s'%str(ddx_p.shape))
        if verbose: even_print('∂y[p]','%s'%str(ddy_p.shape))
    
    if ('data/T' in self):
        
        T = np.copy( self['data/T'][()].T )
        
        if self.rectilinear:
            ddx_T = gradient(T, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_T = gradient(T, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_T_comp = gradient(T, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_T_comp = gradient(T, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_T      = ddx_T_comp*ddx_q1 + ddy_T_comp*ddx_q2
            ddy_T      = ddx_T_comp*ddy_q1 + ddy_T_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_T' in self): del self['data/ddx_T']
        self.create_dataset('data/ddx_T', data=ddx_T.T, chunks=None)
        
        if ('data/ddy_T' in self): del self['data/ddy_T']
        self.create_dataset('data/ddy_T', data=ddy_T.T, chunks=None)
        
        if verbose: even_print('∂x[T]','%s'%str(ddx_T.shape))
        if verbose: even_print('∂y[T]','%s'%str(ddy_T.shape))
    
    if ('data/rho' in self):
        
        rho = np.copy( self['data/rho'][()].T )
        
        if self.rectilinear:
            ddx_rho = gradient(rho, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_rho = gradient(rho, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_r_comp = gradient(rho, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_r_comp = gradient(rho, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_rho    = ddx_r_comp*ddx_q1 + ddy_r_comp*ddx_q2
            ddy_rho    = ddx_r_comp*ddy_q1 + ddy_r_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_rho' in self): del self['data/ddx_rho']
        self.create_dataset('data/ddx_rho', data=ddx_rho.T, chunks=None)
        
        if ('data/ddy_rho' in self): del self['data/ddy_rho']
        self.create_dataset('data/ddy_rho', data=ddy_rho.T, chunks=None)
        
        if verbose: even_print('∂x[ρ]','%s'%str(ddx_rho.shape))
        if verbose: even_print('∂y[ρ]','%s'%str(ddy_rho.shape))
    
    if ('data/mu' in self):
        
        mu = np.copy( self['data/mu'][()].T )
        
        if self.rectilinear:
            ddx_mu = gradient(mu, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_mu = gradient(mu, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_mu_comp = gradient(mu, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_mu_comp = gradient(mu, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_mu      = ddx_mu_comp*ddx_q1 + ddy_mu_comp*ddx_q2
            ddy_mu      = ddx_mu_comp*ddy_q1 + ddy_mu_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_mu' in self): del self['data/ddx_mu']
        self.create_dataset('data/ddx_mu', data=ddx_mu.T, chunks=None)
        
        if ('data/ddy_mu' in self): del self['data/ddy_mu']
        self.create_dataset('data/ddy_mu', data=ddy_mu.T, chunks=None)
        
        if verbose: even_print('∂x[μ]','%s'%str(ddx_mu.shape))
        if verbose: even_print('∂y[μ]','%s'%str(ddy_mu.shape))
    
    # === vorticity
    
    ## z-vorticity :: ω_z
    vort_z = ddx_v - ddy_u
    
    if ('data/vort_z' in self): del self['data/vort_z']
    self.create_dataset('data/vort_z', data=vort_z.T, chunks=None)
    if verbose: even_print('ω_z','%s'%str(vort_z.shape))
    
    ## divergence (in xy-plane)
    div_xy = ddx_u + ddy_v
    
    if ('data/div_xy' in self): del self['data/div_xy']
    self.create_dataset('data/div_xy', data=div_xy.T, chunks=None)
    if verbose: even_print('div_xy','%s'%str(div_xy.shape))
    
    # ==================================================================
    
    if ('data/u_Fv' in self):
        
        u_Fv = np.copy( self['data/u_Fv'][()].T )
        
        if self.rectilinear:
            ddx_u_Fv = gradient(u_Fv, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_u_Fv = gradient(u_Fv, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_u_Fv_comp = gradient(u_Fv, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_u_Fv_comp = gradient(u_Fv, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_u_Fv      = ddx_u_Fv_comp*ddx_q1 + ddy_u_Fv_comp*ddx_q2
            ddy_u_Fv      = ddx_u_Fv_comp*ddy_q1 + ddy_u_Fv_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_u_Fv' in self): del self['data/ddx_u_Fv']
        self.create_dataset('data/ddx_u_Fv', data=ddx_u_Fv.T, chunks=None)
        
        if ('data/ddy_u_Fv' in self): del self['data/ddy_u_Fv']
        self.create_dataset('data/ddy_u_Fv', data=ddy_u_Fv.T, chunks=None)
        
        if verbose: even_print('∂x[u_Fv]','%s'%str(ddx_u_Fv.shape))
        if verbose: even_print('∂y[u_Fv]','%s'%str(ddy_u_Fv.shape))
    
    if ('data/v_Fv' in self):
        
        v_Fv = np.copy( self['data/v_Fv'][()].T )
        
        if self.rectilinear:
            ddx_v_Fv = gradient(v_Fv, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_v_Fv = gradient(v_Fv, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_v_Fv_comp = gradient(v_Fv, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_v_Fv_comp = gradient(v_Fv, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_v_Fv      = ddx_v_Fv_comp*ddx_q1 + ddy_v_Fv_comp*ddx_q2
            ddy_v_Fv      = ddx_v_Fv_comp*ddy_q1 + ddy_v_Fv_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_v_Fv' in self): del self['data/ddx_v_Fv']
        self.create_dataset('data/ddx_v_Fv', data=ddx_v_Fv.T, chunks=None)
        
        if ('data/ddy_v_Fv' in self): del self['data/ddy_v_Fv']
        self.create_dataset('data/ddy_v_Fv', data=ddy_v_Fv.T, chunks=None)
        
        if verbose: even_print('∂x[v_Fv]','%s'%str(ddx_v_Fv.shape))
        if verbose: even_print('∂y[v_Fv]','%s'%str(ddy_v_Fv.shape))
    
    if ('data/utang' in self) and False: ## Retired because could lead to bug and very likely unnecessary
        
        utang = np.copy( self['data/utang'][()].T )
        
        if self.rectilinear:
            ddx_utang = gradient(utang, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_utang = gradient(utang, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_utang_comp = gradient(utang, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_utang_comp = gradient(utang, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_utang      = ddx_utang_comp*ddx_q1 + ddy_utang_comp*ddx_q2
            ddy_utang      = ddx_utang_comp*ddy_q1 + ddy_utang_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_utang' in self): del self['data/ddx_utang']
        self.create_dataset('data/ddx_utang', data=ddx_utang.T, chunks=None)
        if verbose: even_print('∂x[utang]','%s'%str(ddx_utang.shape))
        
        if ('data/ddy_utang' in self): del self['data/ddy_utang']
        self.create_dataset('data/ddy_utang', data=ddy_utang.T, chunks=None)
        if verbose: even_print('∂y[utang]','%s'%str(ddy_utang.shape))
    
    if ('data/unorm' in self) and False: ## Retired because could lead to bug and very likely unnecessary
        
        unorm = np.copy( self['data/unorm'][()].T )
        
        if self.rectilinear:
            ddx_unorm = gradient(unorm, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_unorm = gradient(unorm, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_unorm_comp = gradient(unorm, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_unorm_comp = gradient(unorm, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_unorm      = ddx_unorm_comp*ddx_q1 + ddy_unorm_comp*ddx_q2
            ddy_unorm      = ddx_unorm_comp*ddy_q1 + ddy_unorm_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_unorm' in self): del self['data/ddx_unorm']
        self.create_dataset('data/ddx_unorm', data=ddx_unorm.T, chunks=None)
        if verbose: even_print('∂x[unorm]','%s'%str(ddx_unorm.shape))
        
        if ('data/ddy_unorm' in self): del self['data/ddy_unorm']
        self.create_dataset('data/ddy_unorm', data=ddy_unorm.T, chunks=None)
        if verbose: even_print('∂y[unorm]','%s'%str(ddy_unorm.shape))
    
    if ('data/utang_Fv' in self) and False: ## Retired because could lead to bug and very likely unnecessary
        
        utang_Fv = np.copy( self['data/utang_Fv'][()].T )
        
        if self.rectilinear:
            ddx_utang_Fv = gradient(utang_Fv, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_utang_Fv = gradient(utang_Fv, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_utang_Fv_comp = gradient(utang_Fv, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_utang_Fv_comp = gradient(utang_Fv, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_utang_Fv      = ddx_utang_Fv_comp*ddx_q1 + ddy_utang_Fv_comp*ddx_q2
            ddy_utang_Fv      = ddx_utang_Fv_comp*ddy_q1 + ddy_utang_Fv_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_utang_Fv' in self): del self['data/ddx_utang_Fv']
        self.create_dataset('data/ddx_utang_Fv', data=ddx_utang_Fv.T, chunks=None)
        if verbose: even_print('∂x[utang_Fv]','%s'%str(ddx_utang_Fv.shape))
        
        if ('data/ddy_utang_Fv' in self): del self['data/ddy_utang_Fv']
        self.create_dataset('data/ddy_utang_Fv', data=ddy_utang_Fv.T, chunks=None)
        if verbose: even_print('∂y[utang_Fv]','%s'%str(ddy_utang_Fv.shape))
    
    if ('data/unorm_Fv' in self) and False: ## Retired because could lead to bug and probably unnecessary
        
        unorm_Fv = np.copy( self['data/unorm_Fv'][()].T )
        
        if self.rectilinear:
            ddx_unorm_Fv = gradient(unorm_Fv, self.x, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_unorm_Fv = gradient(unorm_Fv, self.y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        elif self.curvilinear:
            ddx_unorm_Fv_comp = gradient(unorm_Fv, x_comp, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_unorm_Fv_comp = gradient(unorm_Fv, y_comp, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddx_unorm_Fv      = ddx_unorm_Fv_comp*ddx_q1 + ddy_unorm_Fv_comp*ddx_q2
            ddy_unorm_Fv      = ddx_unorm_Fv_comp*ddy_q1 + ddy_unorm_Fv_comp*ddy_q2
        else:
            raise ValueError
        
        if ('data/ddx_unorm_Fv' in self): del self['data/ddx_unorm_Fv']
        self.create_dataset('data/ddx_unorm_Fv', data=ddx_unorm_Fv.T, chunks=None)
        if verbose: even_print('∂x[unorm_Fv]','%s'%str(ddx_unorm_Fv.shape))
        
        if ('data/ddy_unorm_Fv' in self): del self['data/ddy_unorm_Fv']
        self.create_dataset('data/ddy_unorm_Fv', data=ddy_unorm_Fv.T, chunks=None)
        if verbose: even_print('∂y[unorm_Fv]','%s'%str(ddy_unorm_Fv.shape))
    
    # ==================================================================
    # Derivatives in (s,n,z) system : ∂n[],∂s[]
    # ==================================================================
    
    if do_dtangnorm:
        
        print(72*'-')
        
        if self.rectilinear:
            stang = np.copy( self.x )
            snorm = np.copy( self.y )
        elif self.curvilinear:
            stang = np.copy( self.stang )
            snorm = np.copy( self.snorm )
        else:
            raise RuntimeError
        
        quantities = [
            'rho','T','p','mu',
            #'u','v','w', ## ... probably never required
            #'u_Fv','v_Fv', ## ... probably never required
            'utang','unorm','w',
            'utang_Fv','unorm_Fv','w_Fv',
            'utI_utI','unI_unI','utI_unI',
            'r_utII_utII','r_unII_unII','r_utII_unII',
            ]
        
        for Q in quantities:
            dsn_tang = f'data/dds_{Q}' ## Dataset name wall-tangential (s)
            dsn_norm = f'data/ddn_{Q}' ## Dataset name normal (n)
            
            if f'data/{Q}' in self:
                D        = np.copy( self[f'data/{Q}'][()].T ) ## 2D data
                dds_D    = gradient(D, stang, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
                ddn_D    = gradient(D, snorm, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
                
                if (dsn_tang in self):
                    del self[dsn_tang]
                self.create_dataset(dsn_tang, data=dds_D.T, chunks=None)
                if verbose: even_print(f'∂s[{Q}] : {dsn_tang}',str(dds_D.shape))
                
                if (dsn_norm in self):
                    del self[dsn_norm]
                self.create_dataset(dsn_norm, data=ddn_D.T, chunks=None)
                if verbose: even_print(f'∂n[{Q}] : {dsn_norm}',str(ddn_D.shape))
            
            else:
                print(f"WARNING: 'data/{Q}' not found in file.")
                print(f" - cannot compute: ∂s[{Q}]")
                print(f" - cannot compute: ∂n[{Q}]")
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_gradients() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return

def _calc_psvel(self, **kwargs):
    '''
    Calculate pseudovelocity, wall-normal cumulative integration of (-) z-vorticity
    '''
    
    verbose  = kwargs.get('verbose',True)
    
    if verbose: print('\n'+'ztmd.calc_psvel()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if 'data/vort_z' not in self:
        raise ValueError("'data/vort_z' not in ztmd")
    
    vort_z = np.copy( self['data/vort_z'][()].T )
    
    nx = self.nx
    ny = self.ny
    
    if self.rectilinear:
        
        x = np.copy( self['dims/x'][()] )
        y = np.copy( self['dims/y'][()] )
    
    elif self.curvilinear:
        
        ## copy dims into memory
        x = np.copy( self['dims/x'][()].T ) ## 2D
        y = np.copy( self['dims/y'][()].T ) ## 2D
        
        if ('dims/snorm' not in self):
            raise AssertionError('dims/snorm not present')
        if ('dims/stang' not in self):
            raise AssertionError('dims/stang not present')
        
        snorm = np.copy( self['dims/snorm'][()] ) ## 1D
        #stang = np.copy( self['dims/stang'][()] ) ## 1D
        
        if ('data/utang' not in self):
            raise AssertionError('data/utang not present')
        
        #utang = np.copy( self['data/utang'][()].T )
        #unorm = np.copy( self['data/unorm'][()].T )
        
        ## copy csys datasets into memory
        #vtang = np.copy( self['csys/vtang'][()] )
        #vnorm = np.copy( self['csys/vnorm'][()] )
        
        if (x.shape != (self.nx,self.ny)):
            raise ValueError('x.shape != (self.nx,self.ny)')
        if (y.shape != (self.nx,self.ny)):
            raise ValueError('y.shape != (self.nx,self.ny)')
    
    else:
        raise ValueError
    
    ## the local 1D wall-normal coordinate
    if self.rectilinear:
        y_ = np.copy(y)
    elif self.curvilinear:
        y_ = np.copy(snorm)
    else:
        raise ValueError
    
    ## pseudo-velocity is a cumulative integration of (-) z-vorticity
    psvel = np.zeros(shape=(nx,ny), dtype=np.float64)
    for i in range(nx):
        psvel_     = cumulative_trapezoid(-1*vort_z[i,:], y_, initial=0.)
        psvel[i,:] = psvel_
    
    if ('data/psvel' in self):
        del self['data/psvel']
    self.create_dataset('data/psvel', data=psvel.T, chunks=None)
    if verbose: even_print('data/psvel','%s'%str(psvel.shape))
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_psvel() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    
    return

def _calc_wall_quantities(self, acc=6, edge_stencil='full', **kwargs):
    '''
    Get 1D wall quantities
    -----
    - [ ρ_wall, ν_wall, μ_wall, T_wall ]
    - τ_wall = μ_wall·ddn[utang] :: [kg/(m·s)]·[m/s]/[m] = [kg/(m·s²)] = [N/m²] = [Pa]
    - u_τ = (τ_wall/ρ_wall)^(1/2)
    '''
    
    verbose = kwargs.get('verbose',True)
    
    if verbose: print('\n'+'ztmd.calc_wall_quantities()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if verbose: even_print('acc','%i'%(acc,))
    if verbose: even_print('edge_stencil','%s'%(edge_stencil,))
    
    ## check
    if (self.x.ndim==1) and (self.y.ndim==1):
        if hasattr(self,'rectilinear'):
            if not self.rectilinear:
                raise AssertionError
        if hasattr(self,'curvilinear'):
            if self.curvilinear:
                raise AssertionError
    elif (self.x.ndim==2) and (self.y.ndim==2):
        if hasattr(self,'rectilinear'):
            if self.rectilinear:
                raise AssertionError
        if hasattr(self,'curvilinear'):
            if not self.curvilinear:
                raise AssertionError
    else:
        raise ValueError
    
    # ===
    
    if self.curvilinear:
        
        if self.requires_wall_norm_interp:
            
            gndata = 'data_2Dw' ## group name for 2D data (interpolated)
            #gndims = 'dims_2Dw' ## group name for 2D dims (interpolated)
            
            ## wall-normal interpolation coordinates (2D)
            #x_wn = np.copy( self[f'{gndata}/x'][()].T )
            #y_wn = np.copy( self[f'{gndata}/y'][()].T )
            s_wn = np.copy( self[f'{gndata}/wall_distance'][()].T )
        
        else:
            
            gndata = 'data' ## group name for 2D data
            #gndims = 'dims' ## group name for 2D dims
            
            #x_wn = np.copy( self['dims/x'][()].T )
            #y_wn = np.copy( self['dims/y'][()].T )
            s_wn = np.copy( self['dims/snorm'][()] )
            #s_wn = np.broadcast_to(s_wn, (self.nx,self.ny))
        
        if (f'{gndata}/utang' not in self):
            raise AssertionError(f'{gndata}/utang not present')
        
        ## wall-normal interpolated scalars (2D)
        utang_wn  = np.copy( self[f'{gndata}/utang'][()].T  )
        #T_wn      = np.copy( self[f'{gndata}/T'][()].T      )
        vort_z_wn = np.copy( self[f'{gndata}/vort_z'][()].T )
    
    # === get ρ_wall, ν_wall, μ_wall, T_wall
    
    rho = np.copy( self['data/rho'][()].T )
    rho_wall = np.copy( rho[:,0] )
    if ('data_1Dx/rho_wall' in self): del self['data_1Dx/rho_wall']
    self.create_dataset('data_1Dx/rho_wall', data=rho_wall, chunks=None)
    if verbose: even_print('data_1Dx/rho_wall','%s'%str(rho_wall.shape))
    
    nu = np.copy( self['data/nu'][()].T )
    nu_wall = np.copy( nu[:,0] )
    if ('data_1Dx/nu_wall' in self): del self['data_1Dx/nu_wall']
    self.create_dataset('data_1Dx/nu_wall', data=nu_wall, chunks=None)
    if verbose: even_print('data_1Dx/nu_wall','%s'%str(nu_wall.shape))
    
    mu = np.copy( self['data/mu'][()].T )
    mu_wall = np.copy( mu[:,0] )
    if ('data_1Dx/mu_wall' in self): del self['data_1Dx/mu_wall']
    self.create_dataset('data_1Dx/mu_wall', data=mu_wall, chunks=None)
    if verbose: even_print('data_1Dx/mu_wall','%s'%str(mu_wall.shape))
    
    T = np.copy( self['data/T'][()].T )
    T_wall = np.copy( T[:,0] )
    if ('data_1Dx/T_wall' in self): del self['data_1Dx/T_wall']
    self.create_dataset('data_1Dx/T_wall', data=T_wall, chunks=None)
    if verbose: even_print('data_1Dx/T_wall','%s'%str(T_wall.shape))
    
    # === get wall ddn[]
    
    if self.rectilinear:
        
        ddy_u = np.copy( self['data/ddy_u'][()].T )
    
    elif self.curvilinear:
        
        if True:
            
            if (s_wn.ndim==2): ## wall-normal distance (s_norm) is a 2D field
                
                ddn_utang  = np.zeros((self.nx,self.ny), dtype=np.float64) ## dimensional [m/s]/[m] = [1/s]
                ddn_vort_z = np.zeros((self.nx,self.ny), dtype=np.float64)
                
                progress_bar = tqdm(total=self.nx, ncols=100, desc='get ddn[]', leave=False, file=sys.stdout)
                for i in range(self.nx):
                    ddn_utang[i,:]  = gradient(utang_wn[i,:]  , s_wn[i,:], axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
                    ddn_vort_z[i,:] = gradient(vort_z_wn[i,:] , s_wn[i,:], axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
                    progress_bar.update()
                progress_bar.close()
            
            elif (s_wn.ndim==1): ## wall-normal distance (s_norm) is a 1D vector
                
                ddn_utang  = gradient(utang_wn  , s_wn, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
                ddn_vort_z = gradient(vort_z_wn , s_wn, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            
            if (f'{gndata}/ddn_utang' in self): del self[f'{gndata}/ddn_utang']
            self.create_dataset(f'{gndata}/ddn_utang', data=ddn_utang.T, chunks=None)
            
            if (f'{gndata}/ddn_vort_z' in self): del self[f'{gndata}/ddn_vort_z']
            self.create_dataset(f'{gndata}/ddn_vort_z', data=ddn_vort_z.T, chunks=None)
            
            if ('data_1Dx/ddn_utang_wall' in self): del self['data_1Dx/ddn_utang_wall']
            self.create_dataset('data_1Dx/ddn_utang_wall', data=ddn_utang[:,0], chunks=None)
        
        else:
            
            ddn_utang  = np.copy( self[f'{gndata}/ddn_utang'][()].T  )
            ddn_vort_z = np.copy( self[f'{gndata}/ddn_vort_z'][()].T )
    
    else:
        raise ValueError
    
    # === calculate τ_wall & u_τ
    
    ## wall shear stress τ_wall
    if self.rectilinear:
        tau_wall = np.copy( mu_wall * ddy_u[:,0] )
    elif self.curvilinear:
        tau_wall = np.copy( mu_wall * ddn_utang[:,0] )
    else:
        raise ValueError
    
    ## τw
    ## [N/m²] = [kg/(m·s²)] = [Pa]
    if ('data_1Dx/tau_wall' in self): del self['data_1Dx/tau_wall']
    self.create_dataset('data_1Dx/tau_wall', data=tau_wall, chunks=None)
    if verbose: even_print('data_1Dx/tau_wall','%s'%str(tau_wall.shape))
    
    ## friction velocity u_τ [m/s]
    u_tau = np.copy( np.sqrt( tau_wall / rho_wall ) )
    
    if ('data_1Dx/u_tau' in self): del self['data_1Dx/u_tau']
    self.create_dataset('data_1Dx/u_tau', data=u_tau, chunks=None)
    if verbose: even_print('data_1Dx/u_tau','%s'%str(u_tau.shape))
    
    # === inner scales: length, velocity & time
    
    sc_u_in = np.copy( u_tau              )
    sc_l_in = np.copy( nu_wall / u_tau    )
    sc_t_in = np.copy( nu_wall / u_tau**2 )
    np.testing.assert_allclose(sc_t_in, sc_l_in/sc_u_in, rtol=1e-6, atol=1e-12)
    
    if ('data_1Dx/sc_u_in' in self): del self['data_1Dx/sc_u_in']
    self.create_dataset('data_1Dx/sc_u_in', data=sc_u_in, chunks=None)
    if verbose: even_print('data_1Dx/sc_u_in','%s'%str(sc_u_in.shape))
    
    if ('data_1Dx/sc_l_in' in self): del self['data_1Dx/sc_l_in']
    self.create_dataset('data_1Dx/sc_l_in', data=sc_l_in, chunks=None)
    if verbose: even_print('data_1Dx/sc_l_in','%s'%str(sc_l_in.shape))
    
    if ('data_1Dx/sc_t_in' in self): del self['data_1Dx/sc_t_in']
    self.create_dataset('data_1Dx/sc_t_in', data=sc_t_in, chunks=None)
    if verbose: even_print('data_1Dx/sc_t_in','%s'%str(sc_t_in.shape))
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_wall_quantities() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    
    return

def _calc_bl_edge(self, **kwargs):
    '''
    Determine the boundary layer edge location (ZTMD wrapper)
    -----
    'method'=='u'         : 'edge' is where |du+/dy+|<ϵ
    'method'=='vorticity' : 'edge' is where |-ω+|<ϵ
    -----
    y_edge : the wall-normal edge location
    j_edge : the wall-normal edge location grid index
    '''
    
    verbose      = kwargs.get('verbose',True)
    method       = kwargs.get('method','vorticity') ## 'u','vorticity'
    epsilon      = kwargs.get('epsilon',5e-5)
    acc          = kwargs.get('acc',6)
    edge_stencil = kwargs.get('edge_stencil','full')
    interp_kind  = kwargs.get('interp_kind','cubic') ## 'linear','cubic'
    
    if verbose: print('\n'+'ztmd.calc_bl_edge()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    ## Check rectilinear/curvilinear
    if (self.x.ndim==1) and (self.y.ndim==1):
        if hasattr(self,'rectilinear'):
            if not self.rectilinear:
                raise AssertionError
        if hasattr(self,'curvilinear'):
            if self.curvilinear:
                raise AssertionError
    elif (self.x.ndim==2) and (self.y.ndim==2):
        if hasattr(self,'rectilinear'):
            if self.rectilinear:
                raise AssertionError
        if hasattr(self,'curvilinear'):
            if not self.curvilinear:
                raise AssertionError
    else:
        raise ValueError
    
    ## Legacy check
    if hasattr(self,'requires_wall_norm_interp'):
        if self.requires_wall_norm_interp:
            raise NotImplementedError
    
    if not any([(method=='u'),(method=='vorticity')]):
        raise ValueError("'method' should be one of: 'u','vorticity'")
    if not any([(interp_kind=='linear'),(interp_kind=='cubic')]):
        raise ValueError("'interp_kind' should be one of: 'linear','cubic'")
    
    if verbose: even_print('method',method)
    if verbose: even_print('epsilon','%0.1e'%(epsilon,))
    if verbose: even_print('acc',f'{acc:d}')
    if verbose: even_print('edge_stencil',edge_stencil)
    if verbose: even_print('1D interp kind',interp_kind)
    
    # ===
    
    nx = self.nx
    #ny = self.ny
    
    ## copy 2D datasets into memory
    u = np.copy( self['data/u'][()].T ) ## dimensional
    
    u_tau   = np.copy( self['data_1Dx/u_tau'][()] ) ## uτ
    sc_l_in = np.copy( self['data_1Dx/sc_l_in'][()] ) ## δν = νw/uτ
    
    if self.rectilinear:
        
        x = np.copy( self['dims/x'][()] ) ## dimensional
        y = np.copy( self['dims/y'][()] )
    
    elif self.curvilinear:
        
        ## copy dims into memory
        x = np.copy( self['dims/x'][()].T ) ## 2D
        y = np.copy( self['dims/y'][()].T ) ## 2D
        
        if ('dims/snorm' not in self):
            raise AssertionError('dims/snorm not present')
        if ('dims/stang' not in self):
            raise AssertionError('dims/stang not present')
        
        snorm = np.copy( self['dims/snorm'][()] ) ## 1D
        #stang = np.copy( self['dims/stang'][()] ) ## 1D
        
        if ('data/utang' not in self):
            raise AssertionError('data/utang not present')
        
        utang = np.copy( self['data/utang'][()].T )
        #unorm = np.copy( self['data/unorm'][()].T )
        
        ## copy csys datasets into memory
        #vtang = np.copy( self['csys/vtang'][()] )
        vnorm = np.copy( self['csys/vnorm'][()] )
        
        if (x.shape != (self.nx,self.ny)):
            raise ValueError('x.shape != (self.nx,self.ny)')
        if (y.shape != (self.nx,self.ny)):
            raise ValueError('y.shape != (self.nx,self.ny)')
    
    else:
        raise ValueError
    
    # ===
    
    ## The local 1D wall-normal coordinate
    if self.rectilinear:
        y_ = np.copy(y)
    elif self.curvilinear:
        y_ = np.copy(snorm)
    else:
        raise ValueError
    
    # ===
    
    if (method=='vorticity'): ## use -ωz i.e. where |-ωz|<ϵ
        if 'data/vort_z' not in self:
            raise ValueError('data/vort_z not in ztmd')
        vort_z = np.copy( self['data/vort_z'][()].T )
    
    # ===
    
    j_edge    = np.zeros(shape=(nx,)  , dtype=np.int32   )
    y_edge    = np.zeros(shape=(nx,)  , dtype=np.float64 )
    y_edge_2d = np.zeros(shape=(nx,2) , dtype=np.float64 )
    
    if (method=='vorticity'): ## ωz is already the wall-normal derivative of pseudovelocity
        pass
    elif (method=='u'): ## pre-compute du/dy for efficiency
        if self.rectilinear:
            ddy_var = gradient(
                            u,
                            y_,
                            axis=1,
                            d=1,
                            acc=acc,
                            edge_stencil=edge_stencil,
                            )
        elif self.curvilinear:
            ddy_var = gradient(
                            utang,
                            y_,
                            axis=1,
                            d=1,
                            acc=acc,
                            edge_stencil=edge_stencil,
                            )
        else:
            raise ValueError
    else:
        raise ValueError
    
    if verbose: progress_bar = tqdm(total=nx, ncols=100, desc='y_edge', leave=False, file=sys.stdout)
    for i in range(nx):
        
        #do_debug_plot = False
        #if (i==5000):
        #    do_debug_plot = True
        
        if (method=='u'): ## |du+/dy+|<ϵ
            
            ddy_u_      = np.copy( ddy_var[i,:] ) ## du/dy
            ddy_u_plus_ = np.copy( ddy_u_ / ( u_tau[i] / sc_l_in[i] ) ) ## du+/dy+ = (du/dy)/(uτ/δν) = (du/dy)/(uτ^2/νw)
            
            y_edge_, j_edge_ = calc_profile_edge_1d(
                                y=y_,
                                ddy_u=ddy_u_plus_,
                                epsilon=epsilon,
                                )
        
        elif (method=='vorticity'): ## |-ωz+|<ϵ
            
            vort_z_      = np.copy( vort_z[i,:] ) ## ωz = (dv/dx)-(du/dy)
            vort_z_plus_ = np.copy( vort_z_ / ( u_tau[i] / sc_l_in[i] ) ) ## ωz+ = ωz/(uτ/δν) = ωz/(uτ^2/νw)
            
            y_edge_, j_edge_ = calc_profile_edge_1d(
                                y=y_,
                                ddy_u=-1*vort_z_plus_,
                                epsilon=epsilon,
                                )
        
        else:
            raise ValueError
        
        # ===
        
        y_edge[i] = y_edge_
        j_edge[i] = j_edge_
        
        ## 2D [x,y] coordinates of the 'edge line'
        if self.rectilinear:
            pt_edge_ = np.array([self.x[i],y_edge_], dtype=np.float64)
        elif self.curvilinear:
            p0_ = np.array([self.x[i,0],self.y[i,0]], dtype=np.float64)
            vnorm_ = np.copy( vnorm[i,0,:] ) ## unit normal vec @ wall at this x
            pt_edge_ = p0_ + np.dot( y_edge_ , vnorm_ )
        else:
            raise ValueError
        
        y_edge_2d[i,:] = pt_edge_
        
        progress_bar.update()
    progress_bar.close()
    
    if ('data_1Dx/y_edge' in self): del self['data_1Dx/y_edge']
    self.create_dataset('data_1Dx/y_edge', data=y_edge, chunks=None)
    if verbose: even_print('data_1Dx/y_edge','%s'%str(y_edge.shape))
    
    if ('data_1Dx/j_edge' in self): del self['data_1Dx/j_edge']
    self.create_dataset('data_1Dx/j_edge', data=j_edge, chunks=None)
    if verbose: even_print('data_1Dx/j_edge','%s'%str(j_edge.shape))
    
    if ('data_1Dx/y_edge_2d' in self): del self['data_1Dx/y_edge_2d']
    self.create_dataset('data_1Dx/y_edge_2d', data=y_edge_2d, chunks=None)
    if verbose: even_print('data_1Dx/y_edge_2d','%s'%str(y_edge_2d.shape))
    
    # ===
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_bl_edge() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return

def _calc_bl_edge_quantities(self, **kwargs):
    '''
    Calculate field quantity values at [y_edge]
    - Additionally, calculate cf
    '''
    
    verbose = kwargs.get('verbose',True)
    
    if verbose: print('\n'+'ztmd.calc_bl_edge_quantities()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    # ===
    
    nx = self.nx
    ny = self.ny
    
    if self.rectilinear:
        
        ## copy dims into memory
        x = np.copy( self['dims/x'][()] )
        y = np.copy( self['dims/y'][()] )
    
    elif self.curvilinear:
        
        #if ('dims/snorm' not in self):
        #    raise AssertionError('dims/snorm not present')
        #if ('dims/stang' not in self):
        #    raise AssertionError('dims/stang not present')
        
        #snorm = np.copy( self['dims/snorm'][()] ) ## 1D
        #stang = np.copy( self['dims/stang'][()] ) ## 1D
        
        if ('data/utang' not in self):
            raise AssertionError('data/utang not present')
        
        ## copy dims into memory
        x = np.copy( self['dims/x'][()].T )
        y = np.copy( self['dims/y'][()].T )
        
        ## copy csys datasets into memory
        #vtang = np.copy( self['csys/vtang'][()] )
        #vnorm = np.copy( self['csys/vnorm'][()] )
        
        if (x.shape != (self.nx,self.ny)):
            raise ValueError('x.shape != (self.nx,self.ny)')
        if (y.shape != (self.nx,self.ny)):
            raise ValueError('y.shape != (self.nx,self.ny)')
    
    else:
        raise ValueError
    
    #y_edge = np.copy( self['data_1Dx/y_edge'][()]   )
    j_edge = np.copy( self['data_1Dx/j_edge'][()]   )
    
    # ## The local 1D wall-normal coordinate
    # if self.rectilinear:
    #     y_ = np.copy(y)
    # elif self.curvilinear:
    #     y_ = np.copy(snorm)
    # else:
    #     raise ValueError
    
    # === Make a numpy structured array
    
    names  = [ 'rho', 'u', 'v', 'w', 'T', 'p', 'vort_z', 'mu', 'nu', 'M' ]
    if ('data/psvel' in self):
        names += [ 'psvel' ]
    if ('data/utang' in self):
        names += [ 'utang' ]
    if ('data/unorm' in self):
        names += [ 'unorm' ]
    if ('data/umag' in self):
        names += [ 'umag' ]
    
    dtypes=[]
    for n in names:
        ds = self[f'data/{n}']
        dtypes.append( ds.dtype )
    
    names_edge = [ n+'_edge' for n in names ]
    
    data      = np.zeros(shape=(nx,ny), dtype={'names':names,      'formats':dtypes})
    data_edge = np.zeros(shape=(nx,),   dtype={'names':names_edge, 'formats':dtypes})
    
    ## populate 2D structured array with data to find edge for
    for scalar in data.dtype.names:
        data[scalar][:,:] = np.copy( self[f'data/{scalar}'][()].T )
    
    # === Interpolate edge quantity for all vars
    
    if verbose: progress_bar = tqdm(total=nx*len(names), ncols=100, desc='edge quantities', leave=False, file=sys.stdout)
    for scalar in data.dtype.names:
        for i in range(nx):
            je = j_edge[i]
            data_edge_ = data[scalar][i,je]
            data_edge[scalar+'_edge'][i] = data_edge_
            if verbose: progress_bar.update()
    if verbose: progress_bar.close()
    
    # === Write
    
    for scalar in data_edge.dtype.names:
        if (f'data_1Dx/{scalar}' in self):
            del self[f'data_1Dx/{scalar}']
        data_ = np.copy( data_edge[scalar][:] )
        dset = self.create_dataset(f'data_1Dx/{scalar}', data=data_, chunks=None)
        if verbose: even_print(f'data_1Dx/{scalar}',str(dset.shape))
    
    # ===
    
    # if False:
    #     plt.close('all')
    #     fig1 = plt.figure(figsize=(3*2,3), dpi=300)
    #     ax1 = plt.gca()
    #     ax1.plot( stang/self.lchar, data_edge['utang_edge']/self.U_inf, lw=0.5 )
    #     ax1.set_xlabel('stang')
    #     fig1.tight_layout(pad=0.25)
    #     fig1.tight_layout(pad=0.25)
    #     plt.show()
    
    # === Skin-friction coefficient cf = 2·τw/(ρe·ue^2) = 2/(ρe+·(ue+)^2)
    
    tau_wall = np.copy( self['data_1Dx/tau_wall'][()] )
    u_edge   = np.copy( data_edge['u_edge']   )
    rho_edge = np.copy( data_edge['rho_edge'] )
    
    if self.rectilinear:
        cf = np.copy( 2. * tau_wall / ( rho_edge * u_edge**2 ) )
    elif self.curvilinear:
        cf = np.copy( 2. * tau_wall / ( self.rho_inf * self.U_inf**2 ) )
    else:
        raise ValueError
    
    if ('data_1Dx/cf' in self): del self['data_1Dx/cf']
    self.create_dataset('data_1Dx/cf', data=cf, chunks=None)
    if verbose: even_print('data_1Dx/cf', '%s'%str(cf.shape))
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_bl_edge_quantities() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return

def _calc_d99(self, **kwargs):
    '''
    Calculate δ99 (ZTMD wrapper)
    δ = δ99 = y[ u(y) == 0.99*u_edge ]
    'u' can be pseudovelocity or streamwise velocity (set with 'method')
    '''
    
    verbose     = kwargs.get('verbose',True)
    method      = kwargs.get('method','psvel') ## 'u','psvel'
    interp_kind = kwargs.get('interp_kind','cubic') ## 'linear','cubic'
    
    if verbose: print('\n'+'ztmd.calc_d99()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if not any([(method=='u'),(method=='psvel')]):
        raise ValueError("'method' should be one of: 'u','psvel'")
    if not any([(interp_kind=='linear'),(interp_kind=='cubic')]):
        raise ValueError("'interp_kind' should be one of: 'linear','cubic'")
    
    if verbose: even_print('method',method)
    if verbose: even_print('1D interp kind',interp_kind)
    
    # ===
    
    nx = self.nx
    #ny = self.ny
    
    if self.rectilinear:
        
        ## copy dims into memory (1D)
        x = np.copy( self['dims/x'][()] )
        y = np.copy( self['dims/y'][()] )
    
    elif self.curvilinear:
        
        if ('dims/snorm' not in self):
            raise AssertionError('dims/snorm not present')
        if ('dims/stang' not in self):
            raise AssertionError('dims/stang not present')
        
        snorm = np.copy( self['dims/snorm'][()] ) ## 1D
        #stang = np.copy( self['dims/stang'][()] ) ## 1D
        
        if ('data/utang' not in self):
            raise AssertionError('data/utang not present')
        
        ## copy dims into memory (2D)
        x = np.copy( self['dims/x'][()].T )
        y = np.copy( self['dims/y'][()].T )
        
        ## copy csys datasets into memory
        #vtang = np.copy( self['csys/vtang'][()] )
        vnorm = np.copy( self['csys/vnorm'][()] )
        
        if (x.shape != (self.nx,self.ny)):
            raise ValueError('x.shape != (self.nx,self.ny)')
        if (y.shape != (self.nx,self.ny)):
            raise ValueError('y.shape != (self.nx,self.ny)')
    
    else:
        raise ValueError
    
    ## the local 1D wall-normal coordinate
    if self.rectilinear:
        y_ = np.copy(y)
    elif self.curvilinear:
        y_ = np.copy(snorm)
    else:
        raise ValueError
    
    ## The wall normal index of BL edge, as determined in e.g. ztmd.calc_bl_edge()
    j_edge = np.copy( self['data_1Dx/j_edge'][()] )
    
    ## get pseudovelocity / u / utang
    if (method=='psvel'):
        var      = np.copy( self['data/psvel'][()].T )
        var_edge = np.copy( self['data_1Dx/psvel_edge'][()].T )
    elif (method=='u'):
        if self.rectilinear:
            var      = np.copy( self['data/u'][()].T )
            var_edge = np.copy( self['data_1Dx/u_edge'][()] )
        elif self.curvilinear:
            var      = np.copy( self['data/utang'][()].T )
            var_edge = np.copy( self['data_1Dx/utang_edge'][()] )
        else:
            raise ValueError
    else:
        raise ValueError
    
    # ===
    
    d99     = np.zeros(shape=(nx,),  dtype=np.float64 )
    d99_2d  = np.zeros(shape=(nx,2), dtype=np.float64 )
    
    j99     = np.zeros(shape=(nx,),  dtype=np.int32   )
    d99g    = np.zeros(shape=(nx,),  dtype=np.float64 )
    
    if verbose: progress_bar = tqdm(total=nx, ncols=100, desc='δ', leave=False, file=sys.stdout)
    for i in range(nx):
        
        #y_edge_   = y_edge[i]
        j_edge_   = j_edge[i]
        var_      = np.copy( var[i,:] )
        var_edge_ = var_edge[i]
        
        d99_   = calc_d99_1d(y=y_, u=var_, j_edge=j_edge_, u_edge=var_edge_, interp_kind=interp_kind)
        d99[i] = d99_
        
        j99_    = np.abs( y_ - d99_ ).argmin()
        j99[i]  = j99_
        d99g[i] = y_[j99_]
        
        # ===
        
        ## Get the [x,y] coordinates of the 'd99 line' --> shape=(nx,2)
        if self.rectilinear:
            pt_99_ = np.array([self.x[i],d99_], dtype=np.float64)
        elif self.curvilinear:
            p0_ = np.array([self.x[i,0],self.y[i,0]], dtype=np.float64)
            vnorm_ = np.copy( vnorm[i,0,:] ) ## unit normal vec @ wall at this x
            pt_99_ = p0_ + np.dot( d99_ , vnorm_ )
        else:
            raise ValueError
        
        d99_2d[i,:] = pt_99_
        
        progress_bar.update()
    progress_bar.close()
    
    if ('data_1Dx/d99' in self): del self['data_1Dx/d99']
    self.create_dataset('data_1Dx/d99', data=d99, chunks=None)
    if verbose: even_print('data_1Dx/d99','%s'%str(d99.shape))
    
    if ('data_1Dx/d99_2d' in self): del self['data_1Dx/d99_2d']
    self.create_dataset('data_1Dx/d99_2d', data=d99_2d, chunks=None)
    if verbose: even_print('data_1Dx/d99_2d','%s'%str(d99_2d.shape))
    
    if ('data_1Dx/d99g' in self): del self['data_1Dx/d99g']
    self.create_dataset('data_1Dx/d99g', data=d99g, chunks=None)
    if verbose: even_print('data_1Dx/d99g','%s'%str(d99g.shape))
    
    if ('data_1Dx/j99' in self): del self['data_1Dx/j99']
    self.create_dataset('data_1Dx/j99', data=j99, chunks=None)
    if verbose: even_print('data_1Dx/j99','%s'%str(j99.shape))
    
    # ===
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_d99() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return

def _calc_d99_quantities(self, **kwargs):
    '''
    Calculate interpolated field quantity values at y=δ99
    '''
    
    verbose     = kwargs.get('verbose',True)
    interp_kind = kwargs.get('interp_kind','cubic') ## 'linear','cubic'
    
    if verbose: print('\n'+'ztmd.calc_d99_quantities()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if not any([(interp_kind=='linear'),(interp_kind=='cubic')]):
        raise ValueError("'interp_kind' should be one of: 'linear','cubic'")
    
    if verbose: even_print('1D interp kind',interp_kind)
    
    nx = self.nx
    ny = self.ny
    
    if self.rectilinear:
        
        ## copy dims into memory
        x = np.copy( self['dims/x'][()] )
        y = np.copy( self['dims/y'][()] )
    
    elif self.curvilinear:
        
        if ('dims/snorm' not in self):
            raise AssertionError('dims/snorm not present')
        if ('dims/stang' not in self):
            raise AssertionError('dims/stang not present')
        
        snorm = np.copy( self['dims/snorm'][()] ) ## 1D
        #stang = np.copy( self['dims/stang'][()] ) ## 1D
        
        if ('data/utang' not in self):
            raise AssertionError('data/utang not present')
        
        ## copy dims into memory
        x = np.copy( self['dims/x'][()].T )
        y = np.copy( self['dims/y'][()].T )
        
        ## copy csys datasets into memory
        #vtang = np.copy( self['csys/vtang'][()] )
        #vnorm = np.copy( self['csys/vnorm'][()] )
        
        if (x.shape != (self.nx,self.ny)):
            raise ValueError('x.shape != (self.nx,self.ny)')
        if (y.shape != (self.nx,self.ny)):
            raise ValueError('y.shape != (self.nx,self.ny)')
    
    else:
        raise ValueError
    
    d99 = np.copy( self['data_1Dx/d99'][()] )
    
    ## Local 1D wall-normal coordinate
    if self.rectilinear:
        y_ = np.copy(y)
    elif self.curvilinear:
        y_ = np.copy(snorm)
    else:
        raise ValueError
    
    # === Make a structured array
    
    names  = [ 'rho', 'u', 'v', 'w', 'T', 'p', 'vort_z', 'mu', 'nu', 'M' ]
    if ('data/psvel' in self):
        names += [ 'psvel' ]
    if ('data/utang' in self):
        names += [ 'utang' ]
    if ('data/unorm' in self):
        names += [ 'unorm' ]
    if ('data/umag' in self):
        names += [ 'umag' ]
    
    dtypes=[]
    for n in names:
        ds = self[f'data/{n}']
        dtypes.append( ds.dtype )
    
    #names_99 = [ n+'99' for n in names ]
    names_99 = [ n+'_99' if ('_' in n) else n+'99' for n in names ]
    
    data    = np.zeros(shape=(nx,ny), dtype={'names':names,    'formats':dtypes})
    data_99 = np.zeros(shape=(nx,),   dtype={'names':names_99, 'formats':dtypes})
    
    # === populate structured array
    
    for scalar in data.dtype.names:
        data[scalar][:,:] = np.copy( self[f'data/{scalar}'][()].T )
    
    ## Main loop: interpolate @ δ99 for all vars
    if verbose: progress_bar = tqdm(total=nx*len(names), ncols=100, desc='δ99 quantities', leave=False, file=sys.stdout)
    for ni,scalar in enumerate(data.dtype.names):
        for i in range(nx):
            
            data_y_    = np.copy( data[scalar][i,:] )
            intrp_func = interp1d(y_, data_y_, kind=interp_kind, bounds_error=True)
            
            d99_    = d99[i]
            data_99_ = intrp_func(d99_)
            
            data_99[names_99[ni]][i] = data_99_
            
            if verbose: progress_bar.update()
    if verbose: progress_bar.close()
    
    ## Write
    for scalar in data_99.dtype.names:
        if (f'data_1Dx/{scalar}' in self):
            del self[f'data_1Dx/{scalar}']
        data_ = np.copy( data_99[scalar][:] )
        dset = self.create_dataset(f'data_1Dx/{scalar}', data=data_, chunks=None)
        if verbose: even_print(f'data_1Dx/{scalar}',str(dset.shape))
    
    ## Outer (here 99) scales: length, velocity & time
    sc_l_out = np.copy(d99)
    if self.rectilinear:
        sc_u_out = np.copy( data_99['u99'] )
        sc_t_out = np.copy( d99/data_99['u99'] )
    elif self.curvilinear:
        sc_u_out = np.copy( data_99['utang99'] )
        sc_t_out = np.copy( d99/data_99['utang99'] )
    else:
        raise ValueError
    
    np.testing.assert_allclose(sc_t_out, sc_l_out/sc_u_out, rtol=1e-12, atol=1e-12)
    
    u_tau = np.copy( self['data_1Dx/u_tau'][()] )
    sc_t_eddy = np.copy( d99/u_tau )
    
    if ('data_1Dx/sc_u_out' in self): del self['data_1Dx/sc_u_out']
    self.create_dataset('data_1Dx/sc_u_out', data=sc_u_out, chunks=None)
    if verbose: even_print('data_1Dx/sc_u_out', '%s'%str(sc_u_out.shape))
    
    if ('data_1Dx/sc_l_out' in self): del self['data_1Dx/sc_l_out']
    self.create_dataset('data_1Dx/sc_l_out', data=sc_l_out, chunks=None)
    if verbose: even_print('data_1Dx/sc_l_out', '%s'%str(sc_l_out.shape))
    
    if ('data_1Dx/sc_t_out' in self): del self['data_1Dx/sc_t_out']
    self.create_dataset('data_1Dx/sc_t_out', data=sc_t_out, chunks=None)
    if verbose: even_print('data_1Dx/sc_t_out', '%s'%str(sc_t_out.shape))
    
    if ('data_1Dx/sc_t_eddy' in self): del self['data_1Dx/sc_t_eddy']
    self.create_dataset('data_1Dx/sc_t_eddy', data=sc_t_eddy, chunks=None)
    if verbose: even_print('data_1Dx/sc_t_eddy', '%s'%str(sc_t_eddy.shape))
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_d99_quantities() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return

def _calc_bl_integral_quantities(self, **kwargs):
    '''
    Calculate boundary layer integral quantities (ZTMD wrapper)
    δ*=δ1, θ=δ2, Reθ, Reδ2, H12, H32, etc.
    -----
    - Also: Reδ2, Reδ99 out of convenience
    '''
    verbose     = kwargs.get('verbose', True)
    interp_kind = kwargs.get('interp_kind', 'cubic')
    
    if interp_kind not in ('linear', 'cubic'):
        raise ValueError("'interp_kind' should be one of: 'linear','cubic'")
    
    if verbose:
        print('\n' + 'ztmd.calc_bl_integral_quantities()' + '\n' + 72*'-')
    t_start_func = timeit.default_timer()
    
    nx = self.nx
    #ny = self.ny
    
    ## Geometry: Flat-plate vs Curved
    if self.rectilinear:
        y  = np.copy(self['dims/y'][()])
        y_ = y
    elif self.curvilinear:
        y_ = np.copy(self['dims/snorm'][()])
    else:
        raise ValueError
    
    ## 1D [x] profiles
    u_tau    = np.copy(self['data_1Dx/u_tau'][()])
    rho_wall = np.copy(self['data_1Dx/rho_wall'][()])
    nu_wall  = np.copy(self['data_1Dx/nu_wall'][()])
    mu_wall  = np.copy(self['data_1Dx/mu_wall'][()])
    
    #y_edge   = np.copy(self['data_1Dx/y_edge'][()])
    j_edge   = np.copy(self['data_1Dx/j_edge'][()])
    d99      = np.copy(self['data_1Dx/d99'][()])
    rho_edge = np.copy(self['data_1Dx/rho_edge'][()])
    mu_edge  = np.copy(self['data_1Dx/mu_edge'][()])
    nu_edge  = np.copy(self['data_1Dx/nu_edge'][()])
    
    if self.rectilinear:
        u_edge    = np.copy( self['data_1Dx/u_edge'][()])
        #u_Fv_edge = np.copy( self['data_1Dx/u_Fv_edge'][()] )
    else:
        u_edge = np.copy(self['data_1Dx/utang_edge'][()]) ## !! reading 'utang' as local 'u' !!
        #u_Fv_edge = np.copy( self['data_1Dx/utang_Fv_edge'][()] ) ## never tested
    
    ## 2D data
    u   = np.copy(self['data/u'][()].T)
    rho = np.copy(self['data/rho'][()].T)
    
    if self.curvilinear:
        u = np.copy(self['data/utang'][()].T)
    
    ## Output arrays
    d1    = np.full((nx,), np.nan, dtype=np.float64)
    d1_k  = np.full((nx,), np.nan, dtype=np.float64)
    d2    = np.full((nx,), np.nan, dtype=np.float64)
    d2_k  = np.full((nx,), np.nan, dtype=np.float64)
    d3    = np.full((nx,), np.nan, dtype=np.float64)
    d3_k  = np.full((nx,), np.nan, dtype=np.float64)
    dRC   = np.full((nx,), np.nan, dtype=np.float64)
    dRC_k = np.full((nx,), np.nan, dtype=np.float64)
    
    I2   = np.full((nx,), np.nan, dtype=np.float64)
    I2_k = np.full((nx,), np.nan, dtype=np.float64)
    I3   = np.full((nx,), np.nan, dtype=np.float64)
    I3_k = np.full((nx,), np.nan, dtype=np.float64)
    
    if verbose:
        progress_bar = tqdm(total=nx, ncols=100, desc='BL integrals', leave=False)
    
    # Main loop
    # ==================================================================
    
    for i in range(nx):
        
        u_        = u[i,:]
        rho_      = rho[i,:]
        #y_edge_   = y_edge[i]
        j_edge_   = j_edge[i]
        u_edge_   = u_edge[i]
        u_tau_    = u_tau[i]
        rho_edge_ = rho_edge[i]
        rho_wall_ = rho_wall[i]
        
        ## Compressible integrals
        
        #u_edge_ = u_Fv_edge[i]
        
        d1[i] = calc_d1(
            y_, u_,
            rho=rho_,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=rho_edge_,
            )
        
        d2[i] = calc_d2(
            y_, u_,
            rho=rho_,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=rho_edge_,
            )
        
        d3[i] = calc_d3(
            y_, u_,
            rho=rho_,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=rho_edge_,
            )
        
        dRC[i] = calc_dRC(
            y_, u_,
            u_tau=u_tau_,
            rho=rho_,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=rho_edge_,
            rho_wall=rho_wall_,
            )
        
        I2[i] = calc_I2(
            y_, u_,
            u_tau=u_tau_,
            rho=rho_,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=rho_edge_,
            rho_wall=rho_wall_,
            )
        
        I3[i] = calc_I3(
            y_, u_,
            u_tau=u_tau_,
            rho=rho_,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=rho_edge_,
            rho_wall=rho_wall_,
            )
        
        ## Kinetic (u-only) integrals
        
        #u_edge_ = u_Fv_edge[i]
        u_edge_ = u_edge[i]
        
        d1_k[i] = calc_d1(
            y_, u_,
            rho=None,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=None,
            )
        
        d2_k[i] = calc_d2(
            y_, u_,
            rho=None,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=None,
            )
        
        d3_k[i] = calc_d3(
            y_, u_,
            rho=None,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=None,
            )
        
        dRC_k[i] = calc_dRC(
            y_, u_,
            u_tau=u_tau_,
            rho=None,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=None,
            rho_wall=None,
            )
        
        I2_k[i] = calc_I2(
            y_, u_,
            u_tau=u_tau_,
            rho=None,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=None,
            rho_wall=None,
            )
        
        I3_k[i] = calc_I3(
            y_, u_,
            u_tau=u_tau_,
            rho=None,
            j_edge=j_edge_,
            u_edge=u_edge_,
            rho_edge=None,
            rho_wall=None,
            )
        
        if verbose:
            progress_bar.update()
    
    if verbose:
        progress_bar.close()
    
    # ===
    
    ## alias
    dstar   = np.copy( d1   ) ## δ* = δ1
    dstar_k = np.copy( d1_k )
    
    ## alias
    theta   = np.copy( d2   ) ## θ = δ2
    theta_k = np.copy( d2_k )
    
    H12   = np.copy( d1   / d2   ) ## H12 = δ*/θ
    H12_k = np.copy( d1_k / d2_k )
    H32   = np.copy( d3   / d2   ) ## H32 = δ3/δ2
    H32_k = np.copy( d3_k / d2_k )
    
    ## νw = μw / ρw
    np.testing.assert_allclose(
        nu_wall,
        mu_wall / rho_wall,
        rtol=1e-6,
        )
    
    ## νe = μe / ρe
    np.testing.assert_allclose(
        nu_edge,
        mu_edge / rho_edge,
        rtol=1e-6,
        )
    
    ## Reynolds numbers
    Re_tau     = np.copy( d99     * u_tau  * rho_wall / mu_wall ) ## Reτ = δ99·uτ·ρw / μw
    Re_theta   = np.copy( theta   * u_edge * rho_edge / mu_edge ) ## Reθ = θ·ue·ρe / μe
    Re_dstar   = np.copy( dstar   * u_edge * rho_edge / mu_edge ) ## Reδ* = δ*·ue·ρe / μe
    Re_d2      = np.copy( theta   * u_edge * rho_edge / mu_wall ) ## Reδ2 = θ·ue·ρe / μw
    Re_d99     = np.copy( d99     * u_edge * rho_edge / mu_edge ) ## Reδ99 = δ99·ue·ρe / μe
    Re_theta_k = np.copy( theta_k * u_edge * rho_edge / mu_edge ) ## Reθk = θk·ue·ρe / μe
    Re_dstar_k = np.copy( dstar_k * u_edge * rho_edge / mu_edge ) ## Reδ*k = δ*k·ue·ρe / μe
    
    ## Reδ2 = Reθk = (μe/μw)·Reθ ≈ (μ∞/μw)·Reθ
    np.testing.assert_allclose(
        Re_d2,
        (mu_edge / mu_wall) * Re_theta,
        rtol=1e-6,
        )
    
    ## H12 = δ*/θ = Reδ*/Reθ
    np.testing.assert_allclose(
        Re_dstar / Re_theta,
        H12,
        rtol=1e-6,
        )
    
    ## H12k = δ*k/θk = Reδ*k/Reθk
    np.testing.assert_allclose(
        Re_dstar_k / Re_theta_k,
        H12_k,
        rtol=1e-6,
        )
    
    uplus_edge   = np.copy( u_edge   / u_tau    )
    rhoplus_edge = np.copy( rho_edge / rho_wall )
    muplus_edge  = np.copy( mu_edge  / mu_wall  )
    
    ## Reδ*/Reτ = (δ*/δ99)·(ue/uτ)·(ρe/ρw)·(μw/μe)
    ##          = (δ*/δ99)·(ue+·ρe+/μe+)
    np.testing.assert_allclose(
        Re_dstar / Re_tau,
        ( dstar / d99 ) * uplus_edge * rhoplus_edge / muplus_edge,
        rtol=1e-6,
        )
    
    ## Reθ/Reτ = (θ/δ99)·(ue/uτ)·(ρe/ρw)·(μw/μe)
    ##         = (θ/δ99)·(ue+·ρe+/μe+)
    np.testing.assert_allclose(
        Re_theta / Re_tau,
        ( theta / d99 ) * uplus_edge * rhoplus_edge / muplus_edge,
        rtol=1e-6,
        )
    
    ## Helper for writing datasets
    def _write(name, arr):
        if name in self: ## Delete if exists
            del self[name]
        self.create_dataset(name, data=arr, chunks=None) ## Write
        if verbose:
            even_print(name, str(arr.shape))
    
    ## Write datasets to ZTMD HDF5
    _write('data_1Dx/d1'      , d1      )
    _write('data_1Dx/d1_k'    , d1_k    )
    _write('data_1Dx/dstar'   , dstar   )
    _write('data_1Dx/dstar_k' , dstar_k )
    
    _write('data_1Dx/d2'      , d2      )
    _write('data_1Dx/d2_k'    , d2_k    )
    _write('data_1Dx/theta'   , theta   )
    _write('data_1Dx/theta_k' , theta_k )
    
    _write('data_1Dx/d3'    , d3   )
    _write('data_1Dx/d3_k'  , d3_k )
    
    _write('data_1Dx/dRC'   , dRC   )
    _write('data_1Dx/dRC_k' , dRC_k )
    
    _write('data_1Dx/H12'   , H12   )
    _write('data_1Dx/H12_k' , H12_k )
    _write('data_1Dx/H32'   , H32   )
    _write('data_1Dx/H32_k' , H32_k )
    
    _write('data_1Dx/I2'   , I2   )
    _write('data_1Dx/I2_k' , I2_k )
    _write('data_1Dx/I3'   , I3   )
    _write('data_1Dx/I3_k' , I3_k )
    
    _write('data_1Dx/Re_tau'     , Re_tau     )
    _write('data_1Dx/Re_theta'   , Re_theta   )
    _write('data_1Dx/Re_theta_k' , Re_theta_k )
    _write('data_1Dx/Re_d99'     , Re_d99     )
    _write('data_1Dx/Re_d2'      , Re_d2      )
    _write('data_1Dx/Re_dstar'   , Re_dstar   )
    _write('data_1Dx/Re_dstar_k' , Re_dstar_k )
    
    ## Update header
    self.get_header(verbose=False)
    
    if verbose:
        print(72*'-')
        print(
            'total time : ztmd.calc_bl_integral_quantities() : %s'
            % format_time_string(timeit.default_timer() - t_start_func)
            )
        print(72*'-')
    
    return

# ======================================================================

def _calc_wake_parameter(self, **kwargs):
    '''
    Calculate the Coles wake parameter Π (ZTMD wrapper)
    '''
    
    verbose = kwargs.get('verbose',True)
    
    ## Von Kármán constant (κ)
    k = kwargs.get('k',0.384)
    
    ## Log-law intercept (B) : u+ = (1/κ)·ln(y+) + B
    ## see Nagib et al. (2007)
    B = kwargs.get('B',4.173)
    
    if verbose: print('\n'+'ztmd.calc_wake_parameter()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    ## Check
    # ...
    
    if verbose: even_print('κ',f'{k:0.5f}')
    if verbose: even_print('B',f'{B:0.5f}')
    
    if self.curvilinear:
        raise NotImplementedError('ztmd.calc_wake_parameter() has not been implemented for curved cases')
    
    ## Read data
    u_tau   = np.copy( self['data_1Dx/u_tau'][()]   ) ## uτ
    sc_l_in = np.copy( self['data_1Dx/sc_l_in'][()] ) ## δν = νw/uτ
    d99     = np.copy( self['data_1Dx/d99'][()]     ) ## δ99
    u99     = np.copy( self['data_1Dx/u99'][()]     ) ## u(δ99)
    u       = np.copy( self['data/u'][()].T         )
    
    doVD   = False
    doTL   = False
    doVIPL = False
    doGFM  = False
    
    ## Output array
    wake_parameter = np.full((self.nx,),np.nan,dtype=np.float64)
    
    if 'data/u_VD' in self: ## van-Driest
        doVD   = True
        u_VD   = np.copy( self['data/u_VD'][()].T      )
        u99_VD = np.copy( self['data_1Dx/u_99_VD'][()] )
        d99_VD = np.copy( self['data_1Dx/d99_VD'][()]  )
        wake_parameter_VD = np.full((self.nx,),np.nan,dtype=np.float64)
    
    if 'data/u_TL' in self: ## Trettel & Larsson
        doTL   = True
        u_TL   = np.copy( self['data/u_TL'][()].T      )
        y_TL   = np.copy( self['data/y_TL'][()].T      )
        u99_TL = np.copy( self['data_1Dx/u_99_TL'][()] )
        d99_TL = np.copy( self['data_1Dx/d99_TL'][()]  )
        wake_parameter_TL = np.full((self.nx,),np.nan,dtype=np.float64)
    
    if 'data/u_VIPL' in self: ## Volpiani et al.
        doVIPL   = True
        u_VIPL   = np.copy( self['data/u_VIPL'][()].T      )
        y_VIPL   = np.copy( self['data/y_VIPL'][()].T      )
        u99_VIPL = np.copy( self['data_1Dx/u_99_VIPL'][()] )
        d99_VIPL = np.copy( self['data_1Dx/d99_VIPL'][()]  )
        wake_parameter_VIPL = np.full((self.nx,),np.nan,dtype=np.float64)
    
    if 'data/u_GFM' in self: ## Griffin,Fu & Moin
        doGFM   = True
        u_GFM   = np.copy( self['data/u_GFM'][()].T      )
        y_GFM   = np.copy( self['data/y_GFM'][()].T      )
        u99_GFM = np.copy( self['data_1Dx/u_99_GFM'][()] )
        d99_GFM = np.copy( self['data_1Dx/d99_GFM'][()]  )
        wake_parameter_GFM = np.full((self.nx,),np.nan,dtype=np.float64)
    
    if verbose:
        progress_bar = tqdm(
                        total=self.nx,
                        ncols=100,
                        desc='wake parameter',
                        leave=False,
                        file=sys.stdout,
                        )
    for i in range(self.nx):
        
        ## Untransformed
        yplus   = self.y / sc_l_in[i]
        uplus   = u[i,:] / u_tau[i]
        d99plus = d99[i] / sc_l_in[i] ## δ99+
        u99plus = u99[i] / u_tau[i]   ## u+(δ99) for check
        wake_parameter[i] = calc_wake_parameter_1d(yplus=yplus, uplus=uplus, dplus=d99plus, uplus_delta=u99plus, k=k, B=B)
        
        if doVD:
            yplus   = self.y    / sc_l_in[i]
            uplus   = u_VD[i,:] / u_tau[i]
            d99plus = d99_VD[i] / sc_l_in[i]
            u99plus = u99_VD[i] / u_tau[i]
            wake_parameter_VD[i] = calc_wake_parameter_1d(yplus=yplus, uplus=uplus, dplus=d99plus, uplus_delta=u99plus, k=k, B=B)
        
        if doTL:
            yplus   = y_TL[i,:] / sc_l_in[i]
            uplus   = u_TL[i,:] / u_tau[i]
            d99plus = d99_TL[i] / sc_l_in[i]
            u99plus = u99_TL[i] / u_tau[i]
            wake_parameter_TL[i] = calc_wake_parameter_1d(yplus=yplus, uplus=uplus, dplus=d99plus, uplus_delta=u99plus, k=k, B=B)
        
        if doVIPL:
            yplus   = y_VIPL[i,:] / sc_l_in[i]
            uplus   = u_VIPL[i,:] / u_tau[i]
            d99plus = d99_VIPL[i] / sc_l_in[i]
            u99plus = u99_VIPL[i] / u_tau[i]
            wake_parameter_VIPL[i] = calc_wake_parameter_1d(yplus=yplus, uplus=uplus, dplus=d99plus, uplus_delta=u99plus, k=k, B=B)
        
        if doGFM:
            yplus   = y_GFM[i,:] / sc_l_in[i]
            uplus   = u_GFM[i,:] / u_tau[i]
            d99plus = d99_GFM[i] / sc_l_in[i]
            u99plus = u99_GFM[i] / u_tau[i]
            wake_parameter_GFM[i] = calc_wake_parameter_1d(yplus=yplus, uplus=uplus, dplus=d99plus, uplus_delta=u99plus, k=k, B=B)
        
        progress_bar.update()
    progress_bar.close()
    
    ## Helper for writing 1D datasets to data_1Dx/
    def _write_1D(name, arr):
        if name in self: ## Delete if exists
            del self[name]
        self.create_dataset(f'data_1Dx/{name}', data=arr, chunks=None) ## Write
        if verbose:
            even_print(f'data_1Dx/{name}', str(arr.shape))
    
    _write_1D('wake_parameter' , wake_parameter)
    
    if doVD:   _write_1D('wake_parameter_VD'   , wake_parameter_VD   )
    if doTL:   _write_1D('wake_parameter_TL'   , wake_parameter_TL   )
    if doVIPL: _write_1D('wake_parameter_VIPL' , wake_parameter_VIPL )
    if doGFM:  _write_1D('wake_parameter_GFM'  , wake_parameter_GFM  )
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_wake_parameter() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')

def _calc_peak_tauI(self, **kwargs):
    '''
    Calculate peak τ′xx, τ′xy, τ′yy
    '''
    
    verbose = kwargs.get('verbose',True)
    
    if verbose: print('\n'+'ztmd.calc_peak_tauI()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    ## check
    if self.rectilinear:
        pass
    elif self.curvilinear:
        #raise NotImplementedError('ztmd.calc_peak_tauI() has not been implemented for curved cases')
        print('>>> ztmd.calc_peak_tauI() has not been implemented for curved cases')
        return
    else:
        raise ValueError
    
    ## Check data availability
    if 'data_1Dx/sc_l_in' not in self:
        raise ValueError('data_1Dx/sc_l_in not found')
    if 'data/r_uII_uII' not in self:
        raise ValueError('data/r_uII_uII not found')
    if 'data/r_uII_vII' not in self:
        raise ValueError('data/r_uII_vII not found')
    if 'data/r_vII_vII' not in self:
        raise ValueError('data/r_vII_vII not found')
    
    r_uII_uII = np.copy( self['data/r_uII_uII'][()].T )
    r_uII_vII = np.copy( self['data/r_uII_vII'][()].T )
    r_vII_vII = np.copy( self['data/r_vII_vII'][()].T )
    
    y        = np.copy( self['dims/y'][()]            )
    tau_wall = np.copy( self['data_1Dx/tau_wall'][()] )
    sc_l_in  = np.copy( self['data_1Dx/sc_l_in'][()]  )
    sc_l_out = np.copy( self['data_1Dx/sc_l_out'][()] )
    nu_wall  = np.copy( self['data_1Dx/nu_wall'][()]  )
    u_tau    = np.copy( self['data_1Dx/u_tau'][()]    )
    d99      = np.copy( self['data_1Dx/d99'][()]      )
    
    np.testing.assert_allclose(sc_l_in  , nu_wall/u_tau , rtol=1e-12, atol=1e-12)
    np.testing.assert_allclose(sc_l_out , d99           , rtol=1e-12, atol=1e-12)
    
    r_uII_uII_plus = np.copy( r_uII_uII / tau_wall[:,np.newaxis] )
    r_uII_vII_plus = np.copy( r_uII_vII / tau_wall[:,np.newaxis] )
    r_vII_vII_plus = np.copy( r_vII_vII / tau_wall[:,np.newaxis] )
    
    tau_xx_I_peak   = np.zeros((self.nx,), dtype=np.float64)
    tau_xx_I_peak_y = np.zeros((self.nx,), dtype=np.float64)
    
    tau_xy_I_peak   = np.zeros((self.nx,), dtype=np.float64)
    tau_xy_I_peak_y = np.zeros((self.nx,), dtype=np.float64)
    
    tau_yy_I_peak   = np.zeros((self.nx,), dtype=np.float64)
    tau_yy_I_peak_y = np.zeros((self.nx,), dtype=np.float64)
    
    # ===
    
    def __opt_find_peak(y_plus_pk,func):
        root = func(y_plus_pk,1)
        return root
    
    if verbose: progress_bar = tqdm(total=self.nx, ncols=100, desc='peak τ′', leave=False, file=sys.stdout)
    #for i in [5000,10000,15000]:
    for i in range(self.nx):
        
        y_plus_  = np.copy( y / sc_l_in[i]  )
        #yovd_    = np.copy( y / sc_l_out[i] )
        
        r_uII_uII_plus_ = np.copy( r_uII_uII_plus[i,:] )
        r_uII_vII_plus_ = np.copy( r_uII_vII_plus[i,:] )
        r_vII_vII_plus_ = np.copy( r_vII_vII_plus[i,:] )
        
        # === τ′xx
        
        i_naive = np.argmax( r_uII_uII_plus_ )
        #tau_xx_I_peak_naive_ = r_uII_uII_plus_[i_naive]
        
        func_tauIxx = CubicSpline( y_plus_ , r_uII_uII_plus_ , bc_type='natural', extrapolate=False )
        
        bounds_ = ( max(y_plus_[i_naive]*0.9,y_plus_.min()) , min(y_plus_[i_naive]*1.1,y_plus_.max()) )
        
        sol = least_squares(
                fun=__opt_find_peak,
                args=(func_tauIxx,),
                x0=y_plus_[i_naive],
                xtol=1e-15,
                ftol=1e-15,
                gtol=1e-15,
                method='dogbox',
                bounds=bounds_,
                )
        if not sol.success:
            raise ValueError
        
        y_plus_pk_ = float(sol.x[0])
        tau_xx_I_peak[i]   = func_tauIxx(y_plus_pk_) * tau_wall[i] ## re-dimensionalizing
        tau_xx_I_peak_y[i] = y_plus_pk_              * sc_l_in[i]  ## re-dimensionalizing
        
        # ===
        
        # ## debug plot for τ′xx
        # #if (i==5000) or (i==10000) or (i==15000):
        # if 0:
        #     
        #     plt.close('all')
        #     fig1 = plt.figure(figsize=(3,2), dpi=400)
        #     ax1 = plt.gca()
        #     
        #     ax1.tick_params(axis='x', which='both', direction='in')
        #     ax1.tick_params(axis='y', which='both', direction='in')
        #     #ax1.xaxis.set_ticks_position('both')
        #     #ax1.yaxis.set_ticks_position('both')
        #     ax1.set_xscale('log',base=10)
        #     #ax1.set_yscale('log',base=10)
        #     
        #     #ax1.set_xlim(100,3000)
        #     #ax1.xaxis.set_major_locator(mpl.ticker.LogLocator(subs=(1,)))
        #     #ax1.xaxis.set_minor_locator(mpl.ticker.LogLocator(subs=np.linspace(1,9,9)))
        #     #ax1.xaxis.set_minor_formatter(mpl.ticker.NullFormatter())
        #     
        #     ax1.plot(
        #     y_plus_,
        #     r_uII_uII_plus_,
        #     c='k',
        #     zorder=19,
        #     lw=0.8,
        #     marker='o',
        #     ms=1.5,
        #     ls='none',
        #     )
        #     
        #     y_plus_dummy_          = np.logspace(np.log10(1),np.log10(1000),1000)
        #     r_uII_uII_plus_spline_ = func_tauIxx(y_plus_dummy_)
        #     
        #     ax1.plot(
        #     y_plus_dummy_,
        #     r_uII_uII_plus_spline_,
        #     c='blue',
        #     zorder=19,
        #     lw=0.8,
        #     #marker='o',
        #     #ms=1.5,
        #     #ls='none',
        #     )
        #     
        #     ax1.axvline(x=y_plus_pk_       , linestyle='solid', c='gray', zorder=1, lw=0.5)
        #     ax1.axhline(y=func_tauIxx(y_plus_pk_) , linestyle='solid', c='gray', zorder=1, lw=0.5)
        #     
        #     plt.show()
        
        # === τ′xy
        
        i_naive = np.argmin( r_uII_vII_plus_ ) ## ACHTUNG argmin(), NOT argmax() !!
        #tau_xy_I_peak_naive_ = r_uII_vII_plus_[i_naive]
        
        func_tauIxy = CubicSpline( y_plus_ , r_uII_vII_plus_ , bc_type='natural', extrapolate=False )
        
        bounds_ = ( max(y_plus_[i_naive]*0.9,y_plus_.min()) , min(y_plus_[i_naive]*1.1,y_plus_.max()) )
        
        sol = least_squares(
                fun=__opt_find_peak,
                args=(func_tauIxy,),
                x0=y_plus_[i_naive],
                xtol=1e-15,
                ftol=1e-15,
                gtol=1e-15,
                method='dogbox',
                bounds=bounds_,
                )
        if not sol.success:
            raise ValueError
        
        y_plus_pk_ = float(sol.x[0])
        tau_xy_I_peak[i]   = func_tauIxy(y_plus_pk_) * tau_wall[i] ## re-dimensionalizing
        tau_xy_I_peak_y[i] = y_plus_pk_              * sc_l_in[i]  ## re-dimensionalizing
        
        # ===
        
        # ## debug plot for τ′xy
        # #if (i==5000) or (i==10000) or (i==15000):
        # if 0:
        #     
        #     plt.close('all')
        #     fig1 = plt.figure(figsize=(3,2), dpi=400)
        #     ax1 = plt.gca()
        #     
        #     ax1.tick_params(axis='x', which='both', direction='in')
        #     ax1.tick_params(axis='y', which='both', direction='in')
        #     #ax1.xaxis.set_ticks_position('both')
        #     #ax1.yaxis.set_ticks_position('both')
        #     ax1.set_xscale('log',base=10)
        #     #ax1.set_yscale('log',base=10)
        #     
        #     #ax1.set_xlim(100,3000)
        #     #ax1.xaxis.set_major_locator(mpl.ticker.LogLocator(subs=(1,)))
        #     #ax1.xaxis.set_minor_locator(mpl.ticker.LogLocator(subs=np.linspace(1,9,9)))
        #     #ax1.xaxis.set_minor_formatter(mpl.ticker.NullFormatter())
        #     
        #     ax1.plot(
        #     y_plus_,
        #     r_uII_vII_plus_,
        #     c='k',
        #     zorder=19,
        #     lw=0.8,
        #     marker='o',
        #     ms=1.5,
        #     ls='none',
        #     )
        #     
        #     y_plus_dummy_          = np.logspace(np.log10(1),np.log10(1000),1000)
        #     r_uII_vII_plus_spline_ = func_tauIxy(y_plus_dummy_)
        #     
        #     ax1.plot(
        #     y_plus_dummy_,
        #     r_uII_vII_plus_spline_,
        #     c='blue',
        #     zorder=19,
        #     lw=0.8,
        #     #marker='o',
        #     #ms=1.5,
        #     #ls='none',
        #     )
        #     
        #     ax1.axvline(x=y_plus_pk_       , linestyle='solid', c='gray', zorder=1, lw=0.5)
        #     ax1.axhline(y=func_tauIxy(y_plus_pk_) , linestyle='solid', c='gray', zorder=1, lw=0.5)
        #     
        #     plt.show()
        
        # === τ′yy
        
        i_naive = np.argmax( r_vII_vII_plus_ )
        #tau_yy_I_peak_naive_ = r_vII_vII_plus_[i_naive]
        
        func_tauIyy = CubicSpline( y_plus_ , r_vII_vII_plus_ , bc_type='natural', extrapolate=False )
        
        bounds_ = ( max(y_plus_[i_naive]*0.9,y_plus_.min()) , min(y_plus_[i_naive]*1.1,y_plus_.max()) )
        
        sol = least_squares(
                fun=__opt_find_peak,
                args=(func_tauIyy,),
                x0=y_plus_[i_naive],
                xtol=1e-15,
                ftol=1e-15,
                gtol=1e-15,
                method='dogbox',
                bounds=bounds_,
                )
        if not sol.success:
            raise ValueError
        
        y_plus_pk_ = float(sol.x[0])
        tau_yy_I_peak[i]   = func_tauIyy(y_plus_pk_) * tau_wall[i] ## re-dimensionalizing
        tau_yy_I_peak_y[i] = y_plus_pk_              * sc_l_in[i]  ## re-dimensionalizing
        
        # ===
        
        # ## debug plot for τ′yy
        # #if (i==5000) or (i==10000) or (i==15000):
        # if 0:
        #     
        #     plt.close('all')
        #     fig1 = plt.figure(figsize=(3,2), dpi=400)
        #     ax1 = plt.gca()
        #     
        #     ax1.tick_params(axis='x', which='both', direction='in')
        #     ax1.tick_params(axis='y', which='both', direction='in')
        #     #ax1.xaxis.set_ticks_position('both')
        #     #ax1.yaxis.set_ticks_position('both')
        #     ax1.set_xscale('log',base=10)
        #     #ax1.set_yscale('log',base=10)
        #     
        #     #ax1.set_xlim(100,3000)
        #     #ax1.xaxis.set_major_locator(mpl.ticker.LogLocator(subs=(1,)))
        #     #ax1.xaxis.set_minor_locator(mpl.ticker.LogLocator(subs=np.linspace(1,9,9)))
        #     #ax1.xaxis.set_minor_formatter(mpl.ticker.NullFormatter())
        #     
        #     ax1.plot(
        #     y_plus_,
        #     r_vII_vII_plus_,
        #     c='k',
        #     zorder=19,
        #     lw=0.8,
        #     marker='o',
        #     ms=1.5,
        #     ls='none',
        #     )
        #     
        #     y_plus_dummy_          = np.logspace(np.log10(1),np.log10(1000),1000)
        #     r_vII_vII_plus_spline_ = func_tauIyy(y_plus_dummy_)
        #     
        #     ax1.plot(
        #     y_plus_dummy_,
        #     r_vII_vII_plus_spline_,
        #     c='blue',
        #     zorder=19,
        #     lw=0.8,
        #     #marker='o',
        #     #ms=1.5,
        #     #ls='none',
        #     )
        #     
        #     ax1.axvline(x=y_plus_pk_              , linestyle='solid', c='gray', zorder=1, lw=0.5)
        #     ax1.axhline(y=func_tauIyy(y_plus_pk_) , linestyle='solid', c='gray', zorder=1, lw=0.5)
        #     
        #     plt.show()
        
        # ===
        
        progress_bar.update()
    progress_bar.close()
    
    
    if ('data_1Dx/tau_xx_I_peak' in self): del self['data_1Dx/tau_xx_I_peak']
    self.create_dataset('data_1Dx/tau_xx_I_peak', data=tau_xx_I_peak, chunks=None)
    if verbose: even_print('data_1Dx/tau_xx_I_peak','%s'%str(tau_xx_I_peak.shape))
    
    if ('data_1Dx/tau_xy_I_peak' in self): del self['data_1Dx/tau_xy_I_peak']
    self.create_dataset('data_1Dx/tau_xy_I_peak', data=tau_xy_I_peak, chunks=None)
    if verbose: even_print('data_1Dx/tau_xy_I_peak','%s'%str(tau_xy_I_peak.shape))
    
    if ('data_1Dx/tau_yy_I_peak' in self): del self['data_1Dx/tau_yy_I_peak']
    self.create_dataset('data_1Dx/tau_yy_I_peak', data=tau_yy_I_peak, chunks=None)
    if verbose: even_print('data_1Dx/tau_yy_I_peak','%s'%str(tau_yy_I_peak.shape))
    
    # ## old
    # if ('data_1Dx/tau_xx_I_peak_y_plus' in self):
    #     del self['data_1Dx/tau_xx_I_peak_y_plus']
    # if ('data_1Dx/tau_xy_I_peak_y_plus' in self):
    #     del self['data_1Dx/tau_xy_I_peak_y_plus']
    # if ('data_1Dx/tau_yy_I_peak_y_plus' in self):
    #     del self['data_1Dx/tau_yy_I_peak_y_plus']
    
    if ('data_1Dx/tau_xx_I_peak_y' in self): del self['data_1Dx/tau_xx_I_peak_y']
    self.create_dataset('data_1Dx/tau_xx_I_peak_y', data=tau_xx_I_peak_y, chunks=None)
    if verbose: even_print('data_1Dx/tau_xx_I_peak_y','%s'%str(tau_xx_I_peak_y.shape))
    
    if ('data_1Dx/tau_xy_I_peak_y' in self): del self['data_1Dx/tau_xy_I_peak_y']
    self.create_dataset('data_1Dx/tau_xy_I_peak_y', data=tau_xy_I_peak_y, chunks=None)
    if verbose: even_print('data_1Dx/tau_xy_I_peak_y','%s'%str(tau_xy_I_peak_y.shape))
    
    if ('data_1Dx/tau_yy_I_peak_y' in self): del self['data_1Dx/tau_yy_I_peak_y']
    self.create_dataset('data_1Dx/tau_yy_I_peak_y', data=tau_yy_I_peak_y, chunks=None)
    if verbose: even_print('data_1Dx/tau_yy_I_peak_y','%s'%str(tau_yy_I_peak_y.shape))
    
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_peak_tauI() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return



import os
import re
import sys
import timeit
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path, PurePosixPath

import h5py
import numpy as np
import psutil
from mpi4py import MPI
from scipy.signal import csd
from tqdm import tqdm

from .gradient import gradient
from .h5 import h5_print_contents
from .signal import ccor
from .utils import even_print, format_time_string

# ======================================================================

def _calc_wall_coh_xpln(self, **kwargs):
    '''
    Calculate coherence & complex cross-spectrum between turbulent
      field and wall (uτ & τuy) in [t] at every [y]
    ----------------------------------------------------------------
    - Designed for analyzing unsteady, thin planes in [x]
    - Multithreaded with ThreadPoolExecutor()
        - scipy.signal.csd() automatically tries to run multithreaded
        - set OMP_NUM_THREADS=1 and pass 'n_threads' to as kwarg manually
    '''
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    if verbose: print('\n'+'rgd.calc_wall_coh_xpln()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    ## assert that the opened RGD has fsubtype 'unsteady' (i.e. is NOT a prime file)
    if (self.fsubtype!='unsteady'):
        raise ValueError
    if not self.usingmpi:
        raise NotImplementedError('function is not implemented for non-MPI usage')
    
    h5py_is_mpi_build = h5py.h5.get_config().mpi
    if not h5py_is_mpi_build:
        if verbose: print('h5py was not compiled for parallel usage! exiting.')
        sys.exit(1)
    
    rx = kwargs.get('rx',1)
    ry = kwargs.get('ry',1)
    rz = kwargs.get('rz',1)
    rt = kwargs.get('rt',1)
    
    acc          = kwargs.get('acc',6)
    edge_stencil = kwargs.get('edge_stencil','full')
    
    sy = kwargs.get('sy',1) ## N [y] layers to read at a time
    if not isinstance(sy,int) or (sy<1):
        raise TypeError('sy should be a positive non-zero int')
    
    n_threads = kwargs.get('n_threads',1)
    
    ## Debug Rank:Proc Affinity
    #pp = psutil.Process()
    #print(f"[Rank {self.rank}] sees CPUs: {pp.cpu_affinity()}  |  n_threads={n_threads}  |  OMP_NUM_THREADS={os.environ.get('OMP_NUM_THREADS')}")
    
    #try:
    #    n_threads = int(os.environ.get('OMP_NUM_THREADS'))
    #except TypeError: ## not set
    #    n_threads = os.cpu_count()
    
    fn_h5_out       = kwargs.get('fn_h5_out',None)       ## Filename for output HDF5 (.h5) file
    overlap_fac_nom = kwargs.get('overlap_fac_nom',0.50) ## Nominal windows overlap factor
    n_win           = kwargs.get('n_win',8)              ## N segment windows for [t] PSD calc
    
    #overlap_fac_nom = kwargs.get('overlap_fac_nom',0.5)
    #n_win           = kwargs.get('n_win',8)
    
    ## Only distribute data across [y]
    if (rx!=1):
        raise AssertionError('rx!=1')
    if (rz!=1):
        raise AssertionError('rz!=1')
    if (rt!=1):
        raise AssertionError('rt!=1')
    
    if not isinstance(ry,int) or (ry<1):
        raise ValueError('ry should be a positive non-zero int')
    
    ## Check the choice of ranks per dimension
    if (rx*ry*rz*rt != self.n_ranks):
        raise AssertionError('rx*ry*rz*rt != self.n_ranks')
    if (rx>self.nx):
        raise AssertionError('rx>self.nx')
    if (ry>self.ny):
        raise AssertionError('ry>self.ny')
    if (rz>self.nz):
        raise AssertionError('rz>self.nz')
    if (rt>self.nt):
        raise AssertionError('rt>self.nt')
    
    if (self.ny%ry!=0):
        raise ValueError('ny not divisible by ry')
    
    ## Distribute 4D data over ranks --> here only in [y]
    ryl_ = np.array_split(np.arange(self.ny,dtype=np.int64),min(ry,self.ny))
    ryl = [[b[0],b[-1]+1] for b in ryl_ ]
    ry1,ry2 = ryl[self.rank]
    nyr = ry2 - ry1
    
    ## Check all [y] ranges have same size
    for ryl_ in ryl:
        if not (ryl_[1]-ryl_[0]==nyr):
            raise ValueError('[y] chunks are not even in size')
    
    if (nyr%sy!=0):
        raise ValueError('nyr not divisible by sy')
    
    ## Output filename : HDF5 (.h5)
    if (fn_h5_out is None): ## automatically determine name
        fname_path = os.path.dirname(self.fname)
        fname_base = os.path.basename(self.fname)
        fname_root, fname_ext = os.path.splitext(fname_base)
        fname_root = re.findall(r'io\S+_mpi_[0-9]+', fname_root)[0]
        fn_h5_out_base = fname_root+'_coh.h5'
        fn_h5_out = str(PurePosixPath(fname_path, fn_h5_out_base))
    if (Path(fn_h5_out).suffix != '.h5'):
        raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' must end in .h5")
    if os.path.isfile(fn_h5_out):
        #if (os.path.getsize(fn_h5_out) > 8*1024**3):
        #    raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' exists and is >8 [GB]. exiting for your own safety.")
        if (fn_h5_out == self.fname):
            raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' cannot be same as input filename.")
    
    if verbose: even_print( 'fn_h5'      , self.fname )
    if verbose: even_print( 'fn_h5_out'  , fn_h5_out  )
    if verbose: print(72*'-')
    self.comm.Barrier()
    
    ## The data dictionary to be written to .h5 later
    data = {}
    
    ## Infile
    fsize = os.path.getsize(self.fname)/1024**3
    if verbose: even_print(os.path.basename(self.fname),'%0.1f [GB]'%fsize)
    if verbose: even_print('nx',f'{self.nx:d}')
    if verbose: even_print('ny',f'{self.ny:d}')
    if verbose: even_print('nz',f'{self.nz:d}')
    if verbose: even_print('nt',f'{self.nt:d}')
    if verbose: even_print('ngp',f'{self.ngp/1e6:0.1f} [M]')
    #if verbose: even_print('cy',f'{cy:d}')
    if verbose: even_print('sy',f'{sy:d}')
    if verbose: even_print('n_ranks',f'{self.n_ranks:d}')
    if verbose: even_print('n_threads',f'{n_threads:d}')
    if verbose: print(72*'-')
    
    ## 0D freestream scalars
    lchar   = self.lchar   ; data['lchar']   = lchar
    U_inf   = self.U_inf   ; data['U_inf']   = U_inf
    rho_inf = self.rho_inf ; data['rho_inf'] = rho_inf
    T_inf   = self.T_inf   ; data['T_inf']   = T_inf
    
    #data['M_inf'] = self.M_inf
    data['Ma'] = self.Ma
    data['Pr'] = self.Pr
    
    ## Read in 1D coordinate arrays & re-dimensionalize
    x = np.copy( self['dims/x'][()] * self.lchar )
    y = np.copy( self['dims/y'][()] * self.lchar )
    z = np.copy( self['dims/z'][()] * self.lchar )
    t = np.copy( self['dims/t'][()] * self.tchar )
    
    nx = self.nx ; data['nx'] = nx
    ny = self.ny ; data['ny'] = ny
    nz = self.nz ; data['nz'] = nz
    nt = self.nt ; data['nt'] = nt
    
    ## Assert constant Δz
    dz0 = np.diff(z)[0]
    if not np.all(np.isclose(np.diff(z), dz0, rtol=1e-6)):
        raise NotImplementedError('Δz not constant')
    dz = np.diff(z)[0]
    
    ## dimensional [s]
    dt = self.dt * self.tchar
    np.testing.assert_allclose(dt, t[1]-t[0], rtol=1e-12, atol=1e-12)
    
    t_meas = self.duration * self.tchar
    np.testing.assert_allclose(t_meas, t.max()-t.min(), rtol=1e-12, atol=1e-12)
    
    zrange = z.max() - z.min()
    
    data['x'] = x
    data['y'] = y
    data['z'] = z
    
    data['t']      = t
    data['t_meas'] = t_meas
    data['dt']     = dt
    data['dz']     = dz
    data['zrange'] = zrange
    
    if verbose: even_print( 'Δt/tchar'       , f'{dt/self.tchar:0.8f}' )
    if verbose: even_print( 'Δt'             , f'{dt:0.3e} [s]'        )
    if verbose: even_print( 'duration/tchar' , f'{self.duration:0.1f}' )
    if verbose: even_print( 'duration'       , f'{self.duration*self.tchar:0.3e} [s]' )
    if verbose: print(72*'-')
    
    ## report
    if verbose:
        even_print('Δt'     , f'{dt    :0.5e} [s]' )
        even_print('t_meas' , f'{t_meas:0.5e} [s]' )
        even_print('Δz'     , f'{dz0   :0.5e} [m]' )
        even_print('zrange' , f'{zrange:0.5e} [m]' )
        print(72*'-')
    
    ## Establish [t] windowing & get frequency
    nperseg     = nt // n_win
    noverlap    = int(round(nperseg*overlap_fac_nom))
    overlap_fac = noverlap / nperseg
    fs          = 1./dt ## dimensional [1/s]
    
    ## Get [freq] vector
    freq_w0,_ = csd(
            np.zeros((nt,),dtype=np.float64),
            np.zeros((nt,),dtype=np.float64),
            fs=fs,
            nperseg=nperseg,
            noverlap=noverlap,
            window='hann',
            detrend=False,
            scaling='density',
            return_onesided=True,
            )
    fp   = np.where(freq_w0>0) ## dont include 0 freq
    freq = np.copy(freq_w0[fp])
    nf   = freq.shape[0]
    df   = np.diff(freq)[0]
    
    data['nperseg']  = nperseg
    data['noverlap'] = noverlap
    data['freq']     = freq
    data['df']       = df
    data['nf']       = nf
    
    if verbose:
        even_print('overlap_fac (nominal)' , f'{overlap_fac_nom:0.5f}'  )
        even_print('n_win'                 , f'{n_win:d}'               )
        even_print('nperseg'               , f'{nperseg:d}'             )
        even_print('noverlap'              , f'{noverlap:d}'            )
        even_print('overlap_fac'           , f'{overlap_fac:0.5f}'      )
        print(72*'-')
    
    if verbose:
        even_print('freq min',f'{freq.min():0.1f} [Hz]')
        even_print('freq max',f'{freq.max():0.1f} [Hz]')
        even_print('df',f'{df:0.1f} [Hz]')
        even_print('nf',f'{nf:d}')
        print(72*'-')
    
    ## ## [z] Wavenumber (kz) vector -- scipy fftfreq version
    ## kz_full = sp.fft.fftfreq(n=nz, d=dz0) * ( 2 * np.pi )
    ## kzp     = np.where(kz_full>0) ## dont include k=0 or (-) k
    ## kz      = np.copy(kz_full[kzp])
    ## dkz     = kz[1] - kz[0]
    ## nkz     = kz.shape[0]
    
    ## [z] Wavenumber (kz) vector
    kz_ov_2pi,_ = csd(
                    np.zeros((nz,),dtype=np.float64),
                    np.zeros((nz,),dtype=np.float64),
                    fs=1/dz0,
                    nperseg=nz,
                    noverlap=0,
                    window='boxcar',
                    detrend=False,
                    scaling='density',
                    return_onesided=True,
                    )
    kz_full = kz_ov_2pi * (2 * np.pi)
    kzp     = np.where(kz_full>0) ## dont include k=0
    kz      = np.copy(kz_full[kzp])
    dkz = kz[1] - kz[0]
    nkz = kz.shape[0]
    
    data['kz']  = kz
    data['dkz'] = dkz
    data['nkz'] = nkz
    
    if verbose:
        even_print('kz min',f'{kz.min():0.1f} [1/m]')
        even_print('kz max',f'{kz.max():0.1f} [1/m]')
        even_print('dkz',f'{dkz:0.1f} [1/m]')
        even_print('nkz',f'{nkz:d}')
        print(72*'-')
    
    ## Wavelength λz = (2·π)/kz
    lz = np.copy( 2 * np.pi / kz )
    data['lz'] = lz
    
    # ===
    
    ## Get lags [t]
    lags_t,_  = ccor( np.ones(nt,dtype=np.float32) , np.ones(nt,dtype=np.float32), get_lags=True )
    n_lags_t_ = nt*2-1
    n_lags_t  = lags_t.shape[0]
    if (n_lags_t!=n_lags_t_):
        raise AssertionError('check lags [t]')
    
    data['lags_t']   = lags_t
    data['n_lags_t'] = n_lags_t
    
    if verbose:
        even_print('n lags (Δt)' , '%i'%(n_lags_t,))
    
    # ===
    
    ## cross-correlation pairs
    ## [ str:var1, str:var2, bool:do_density_weighting]
    ccor_combis = [
        
        [ 'utau'   , 'u' , True  ], ## [ uτ′  , ρ·u″ ]
        [ 'utau'   , 'v' , True  ], ## [ uτ′  , ρ·v″ ]
        [ 'utau'   , 'u' , False ], ## [ uτ′  , u′   ]
        [ 'utau'   , 'v' , False ], ## [ uτ′  , v′   ]
        [ 'utau'   , 'p' , False ], ## [ uτ′  , p′   ]
        [ 'utau'   , 'T' , False ], ## [ uτ′  , T′   ]
        
        [ 'tauuy'  , 'u' , True  ], ## [ τuy′ , ρ·u″ ]
        [ 'tauuy'  , 'v' , True  ], ## [ τuy′ , ρ·v″ ]
        [ 'tauuy'  , 'u' , False ], ## [ τuy′ , u′   ]
        [ 'tauuy'  , 'v' , False ], ## [ τuy′ , v′   ]
        [ 'tauuy'  , 'p' , False ], ## [ τuy′ , p′   ]
        [ 'tauuy'  , 'T' , False ], ## [ τuy′ , T′   ]
        
        ]
    
    ## Generate cross-correlation scalar names
    scalars = []
    for ccor_combi in ccor_combis:
        s1,s2,do_density_weighting = ccor_combi
        if do_density_weighting:
            scalars.append(f'{s1}I_r{s2}II')
        else:
            scalars.append(f'{s1}I_{s2}I')
    
    ## Generate avg scalar names
    scalars_Re_avg = []
    scalars_Fv_avg = []
    for ccor_combi in ccor_combis:
        s1,s2,do_density_weighting = ccor_combi
        if do_density_weighting and ('rho' not in scalars_Re_avg):
            scalars_Re_avg.append('rho')
        if do_density_weighting:
            #if (s1 not in scalars_Fv_avg):
            #    scalars_Fv_avg.append(s1)
            if (s2 not in scalars_Fv_avg):
                scalars_Fv_avg.append(s2)
        else:
            #if (s1 not in scalars_Re_avg):
            #    scalars_Re_avg.append(s1)
            if (s2 not in scalars_Re_avg):
                scalars_Re_avg.append(s2)
    
    ## numpy formatted arrays: buffers for PSD & other data (rank-local)
    Rt         = np.zeros(shape=(nyr, n_lags_t ) , dtype={'names':scalars        , 'formats':[np.dtype(np.float64)    for s in scalars]})
    Coh_t      = np.zeros(shape=(nyr, nf       ) , dtype={'names':scalars        , 'formats':[np.dtype(np.complex128) for s in scalars]})
    Coh_z      = np.zeros(shape=(nyr, nkz      ) , dtype={'names':scalars        , 'formats':[np.dtype(np.complex128) for s in scalars]})
    Pt         = np.zeros(shape=(nyr, nf       ) , dtype={'names':scalars        , 'formats':[np.dtype(np.complex128) for s in scalars]})
    Pz         = np.zeros(shape=(nyr, nkz      ) , dtype={'names':scalars        , 'formats':[np.dtype(np.complex128) for s in scalars]})
    covariance = np.zeros(shape=(nyr,          ) , dtype={'names':scalars        , 'formats':[np.dtype(np.float64)    for s in scalars]})
    avg_Re     = np.zeros(shape=(nyr,          ) , dtype={'names':scalars_Re_avg , 'formats':[np.dtype(np.float64)    for s in scalars_Re_avg]})
    avg_Fv     = np.zeros(shape=(nyr,          ) , dtype={'names':scalars_Fv_avg , 'formats':[np.dtype(np.float64)    for s in scalars_Fv_avg]})
    
    if verbose:
        even_print('n cross-correlation scalar combinations' , f'{len(ccor_combis):d}')
        print(72*'-')
    
    # ==============================================================
    # Calculate instantaneous uτ & τuy
    # ==============================================================
    
    if verbose:
        print('>>> calculating uτ & τuy')
    self.comm.Barrier()
    t_start = timeit.default_timer()
    
    if (self.rank==0):
        
        u        = np.zeros(shape=(nx,7,nz,nt), dtype=np.float64)
        T_wall   = np.zeros(shape=(nx,1,nz,nt), dtype=np.float64)
        rho_wall = np.zeros(shape=(nx,1,nz,nt), dtype=np.float64)
        
        dset = self['data/u']
        #with dset.collective:
        u[:,:,:,:] = dset[:,:,:7,:].T
        dset = self['data/T']
        #with dset.collective:
        T_wall[:,:,:,:] = dset[:,:,0,:].T[:,np.newaxis,:,:]
        dset = self['data/rho']
        #with dset.collective:
        rho_wall[:,:,:,:] = dset[:,:,0,:].T[:,np.newaxis,:,:]
        
        ## Re-dimensionalize
        u        *= self.U_inf
        T_wall   *= self.T_inf
        rho_wall *= self.rho_inf
        
        mu_wall          = np.zeros(shape=(nx,1,nz,nt), dtype=np.float64)
        mu_wall[:,:,:,:] = self.mu_Suth_ref * ( T_wall / self.T_Suth_ref )**(3/2) * ( ( self.T_Suth_ref + self.S_Suth ) / ( T_wall + self.S_Suth ) )
        
        ddy_u          = np.zeros(shape=(nx,7,nz,nt), dtype=np.float64)
        ddy_u[:,:,:,:] = gradient(u, y[:7], axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        
        ddy_u_wall          = np.zeros(shape=(nx,1,nz,nt), dtype=np.float64)
        ddy_u_wall[:,:,:,:] = ddy_u[:,0,:,:][:,np.newaxis,:,:]
        
        ddy_u = None ; del ddy_u
        u     = None ; del u
        
        ## INSTANTANEOUS τw
        tau_uy          = np.zeros(shape=(nx,1,nz,nt), dtype=np.float64)
        tau_uy[:,:,:,:] = mu_wall[:,:,:,:] * ddy_u_wall[:,:,:,:]
        
        ## INSTANTANEOUS uτ
        u_tau          = np.zeros(shape=(nx,1,nz,nt), dtype=np.float64)
        u_tau[:,:,:,:] = np.sign(tau_uy) * np.sqrt( np.abs(tau_uy) / rho_wall )
        
        mu_wall  = None ; del mu_wall
        T_wall   = None ; del T_wall
        rho_wall = None ; del rho_wall
        
        if ( u_tau.shape != (nx,1,nz,nt) ) or ( tau_uy.shape != (nx,1,nz,nt) ):
            print(f'rank {self.rank:d}: shape violation')
            self.comm.Abort(1)
        
        u_tau_avg  = np.mean(u_tau  , axis=3, dtype=np.float64, keepdims=True) ## (x,1,z,1)
        tau_uy_avg = np.mean(tau_uy , axis=3, dtype=np.float64, keepdims=True) ## (x,1,z,1)
        
        if ( u_tau_avg.shape != (nx,1,nz,1) ) or ( tau_uy_avg.shape != (nx,1,nz,1) ):
            print(f'rank {self.rank:d}: shape violation')
            self.comm.Abort(1)
        
        u_tau_avg  = None ; del u_tau_avg
        tau_uy_avg = None ; del tau_uy_avg
    
    # ==============================================================
    
    self.comm.Barrier()
    t_delta = timeit.default_timer() - t_start
    if verbose:
        even_print('calculate uτ & τuy',format_time_string(t_delta))
        print(72*'-')
    
    ## Initialize buffers on non-0 ranks
    if self.rank!=0:
        tau_uy = np.zeros(shape=(nx,1,nz,nt), dtype=np.float64)
        u_tau  = np.zeros(shape=(nx,1,nz,nt), dtype=np.float64)
    
    ## Broadcast from rank 0 to all ranks
    self.comm.Barrier()
    t_start = timeit.default_timer()
    
    self.comm.Bcast( tau_uy , root=0 )
    self.comm.Bcast( u_tau  , root=0 )
    
    self.comm.Barrier()
    t_delta = timeit.default_timer() - t_start
    
    if verbose:
        even_print('Bcast uτ & τuy',format_time_string(t_delta))
        print(72*'-')
    
    # ==============================================================
    # Check memory
    # ==============================================================
    
    hostname = MPI.Get_processor_name()
    mem_free_gb = psutil.virtual_memory().free / 1024**3
    G = self.comm.gather([ self.rank , hostname , mem_free_gb ], root=0)
    G = self.comm.bcast(G, root=0)
    
    host_mem = {}
    for rank, host, mem in G:
        if host not in host_mem or mem < host_mem[host]:
            host_mem[host] = mem
    total_free = sum(host_mem.values())
    
    if verbose:
        #print(72*'-')
        for key,value in host_mem.items():
            even_print(f'RAM free {key}', f'{int(np.floor(value)):d} [GB]')
        even_print('RAM free (local,min)', f'{int(np.floor(min(host_mem.values()))):d} [GB]')
        even_print('RAM free (global)', f'{int(np.floor(total_free)):d} [GB]')
    
    shape_read = (nx,sy,nz,nt) ## local
    if verbose: even_print('read shape (local)', f'[{nx:d},{sy:d},{nz:d},{nt:d}]')
    data_gb = np.dtype(np.float64).itemsize * np.prod(shape_read) / 1024**3
    if verbose: even_print('read size (global)', f'{int(np.ceil(data_gb*ry)):d} [GB]')
    
    if verbose: even_print('read size (global) ×6', f'{int(np.ceil(data_gb*ry*6)):d} [GB]')
    ram_usage_est = data_gb*ry*6/total_free
    if verbose: even_print('RAM usage estimate', f'{100*ram_usage_est:0.1f} [%]')
    
    self.comm.Barrier()
    if (ram_usage_est>0.80):
        print('RAM consumption might be too high. exiting.')
        self.comm.Abort(1)
    
    # ==============================================================
    # Main loop
    # ==============================================================
    
    if verbose:
        progress_bar = tqdm(
            #total=len(ccor_combis)*cy,
            total=len(ccor_combis)*(nyr//sy),
            ncols=100,
            desc='Coh',
            leave=True,
            file=sys.stdout,
            mininterval=0.1,
            smoothing=0.,
            #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
            bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
            ascii="░█",
            colour='#FF6600',
            )
    
    for cci,cc in enumerate(ccor_combis): ## ccor pairs
        
        if verbose: tqdm.write(72*'-')
        
        scalar_L, scalar_R, do_density_weighting = cc
        
        if scalar_L == 'utau':
            scalar_L_str = 'uτ'
        elif scalar_L == 'tauuy':
            scalar_L_str = 'τuy'
        else:
            raise RuntimeError
        
        if do_density_weighting:
            msg = f'[{scalar_L_str}′,ρ·{scalar_R}″]'
        else:
            msg = f'[{scalar_L_str}′,{scalar_R}′]'
        if verbose:
            tqdm.write(even_print('computing',msg,s=True,))
        
        #dset_L   = self[f'data/{scalar_L}']
        dset_R   = self[f'data/{scalar_R}']
        dset_rho = self['data/rho']
        
        scalar = scalars[cci]
        
        ## Assert scalar name
        if do_density_weighting:
            if (f'{scalar_L}I_r{scalar_R}II' != scalar ):
                raise ValueError
        else:
            if (f'{scalar_L}I_{scalar_R}I' != scalar ):
                raise ValueError
        
        # ## [y] loop outer (chunks within rank)
        # for cyl_ in cyl:
        #     cy1, cy2 = cyl_
        #     nyc = cy2 - cy1
        
        for ci in range(nyr//sy): ## [y] loop
            
            cy1 = ry1 + ci*sy
            cy2 = cy1 + sy
            nyc = cy2 - cy1
            
            ## COPY data L (no read!)
            if scalar_L == 'utau':
                data_L = np.zeros(shape=(nx,1,nz,nt), dtype=np.float64)
                data_L[:,:,:,:] = u_tau[:,:,:,:]
            elif scalar_L == 'tauuy':
                data_L = np.zeros(shape=(nx,1,nz,nt), dtype=np.float64)
                data_L[:,:,:,:] = tau_uy[:,:,:,:]
            else:
                raise RuntimeError
            
            self.comm.Barrier()
            t_start = timeit.default_timer()
            
            ## Read data R
            scalar_str = scalar_R
            n_scalars_read = 1
            with dset_R.collective:
                data_R = np.copy( dset_R[:,:,cy1:cy2,:].T ).astype(np.float64)
            
            ## Read ρ
            if do_density_weighting:
                n_scalars_read += 1
                scalar_str += ',ρ'
                with dset_rho.collective:
                    rho = np.copy( dset_rho[:,:,cy1:cy2,:].T ).astype(np.float64)
            else:
                rho = None
            
            self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            data_gb = n_scalars_read * ( self.nx * ry * (cy2-cy1) * self.nz * self.nt * dset_R.dtype.itemsize ) / 1024**3
            if verbose:
                tqdm.write(even_print(f'read: {scalar_str}', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
            
            ## Assert shapes
            if ( data_L.shape != (nx,1,nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            if ( data_R.shape != (nx,nyc,nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            if (rho is not None) and ( rho.shape != (nx,nyc,nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            # === Redimensionalize
            
            if scalar_R in ['u','v','w',]:
                data_R *= U_inf
            elif scalar_R in ['p',]:
                data_R *= rho_inf*U_inf**2
            elif scalar_R in ['T',]:
                data_R *= T_inf
            else:
                raise ValueError
            
            if (rho is not None): ## i.e. if do_density_weighting
                rho *= rho_inf
            
            # === Compute mean-removed data
            
            ## avg(□) or avg(ρ·□)/avg(ρ) in [t]
            if do_density_weighting:
                rho_avg     = np.mean(        rho , axis=3, dtype=np.float64, keepdims=True) ## [x,y,z,1]
                data_R_avg  = np.mean( rho*data_R , axis=3, dtype=np.float64, keepdims=True) ## [x,y,z,1]
                data_R_avg /= rho_avg
            else:
                data_R_avg = np.mean( data_R , axis=3, dtype=np.float64, keepdims=True) ## [x,y,z,1]
            
            ### pointer to data L
            #if scalar_L == 'utau':
            #    data_L_avg = u_tau_avg
            #elif scalar_L == 'tauuy':
            #    data_L_avg = tau_uy_avg
            #else:
            #    raise RuntimeError
            
            data_L_avg = np.mean( data_L , axis=3, dtype=np.float64, keepdims=True) ## (x,y,z,1)
            
            ## Reynolds prime □′ or Favre prime □″
            data_L -= data_L_avg
            data_R -= data_R_avg
            
            ## Assert stationarity / definition averaging
            ## avg(□′)==0 or avg(ρ·□″)==0
            if do_density_weighting:
                b_ = np.mean(rho*data_R, axis=3, dtype=np.float64, keepdims=True)
            else:
                b_ = np.mean(data_R, axis=3, dtype=np.float64, keepdims=True)
            if not np.allclose( b_, np.zeros_like(b_), atol=1e-6 ):
                print(f'rank {self.rank:d}: avg(□′)!=0 or avg(ρ·□″)!=0')
                self.comm.Abort(1)
            
            ## LEFT variable (uτ′ or τuy′) ... never gets ρ-weighted / Favre mean-removed
            a_ = np.mean(data_L, axis=3, dtype=np.float64, keepdims=True)
            if not np.allclose( a_, np.zeros_like(a_), atol=1e-6 ):
                print(f'rank {self.rank:d}: avg(□′)!=0 or avg(ρ·□″)!=0')
                self.comm.Abort(1)
            
            a_ = None ; del a_
            b_ = None ; del b_
            
            ## Covariance (L here is never ρ-weighted)
            if do_density_weighting:
                covariance_ = np.mean( data_L * rho*data_R , axis=3 , dtype=np.float64, keepdims=True)
            else:
                covariance_ = np.mean( data_L * data_R , axis=3 , dtype=np.float64, keepdims=True)
            
            ## Write this chunk/scalar's covariance to covariance buffer
            ## avg over [x,z] : [x,y,z,1] --> [y]
            yiA = cy1 - ry1
            yiB = cy2 - ry1
            covariance[scalar][yiA:yiB] = np.squeeze( np.mean( covariance_ , axis=(0,2,3) , dtype=np.float64) )
            
            ## Write (rank-local) 1D [y] averages to buffers
            if do_density_weighting:
                avg_Fv[scalar_R][yiA:yiB] = np.squeeze( np.mean( data_R_avg , axis=(0,2,3) , dtype=np.float64) )
                avg_Re['rho'][yiA:yiB]    = np.squeeze( np.mean( rho_avg    , axis=(0,2,3) , dtype=np.float64) )
            else:
                avg_Re[scalar_R][yiA:yiB] = np.squeeze( np.mean( data_R_avg , axis=(0,2,3) , dtype=np.float64) )
            
            # ===============================================================================
            # At this point you have 4D [x,y,z,t] [_,□′] or [_,ρ·□″] data
            # ===============================================================================
            
            def __ccor_kernel_t(xi,zi,yii,do_density_weighting):
                if do_density_weighting:
                    uR = rho[xi,yii,zi,:] * data_R[xi,yii,zi,:]
                else:
                    uR = data_R[xi,yii,zi,:]
                uL = data_L[xi,0,zi,:]
                return xi,zi,ccor(uL,uR)
            
            def __coherence_kernel_t(xi,zi,yii,do_density_weighting):
                
                ## 1D [t] □′ or ρ·□″ vectors
                if do_density_weighting:
                    uR = rho[xi,yii,zi,:] * data_R[xi,yii,zi,:]
                else:
                    uR = data_R[xi,yii,zi,:]
                uL = data_L[xi,0,zi,:]
                
                _,Pxx = csd(
                        uL,uL,
                        fs=fs,
                        nperseg=nperseg,
                        noverlap=noverlap,
                        window='hann',
                        detrend=False,
                        scaling='density',
                        return_onesided=True,
                        )
                _,Pyy = csd(
                        uR,uR,
                        fs=fs,
                        nperseg=nperseg,
                        noverlap=noverlap,
                        window='hann',
                        detrend=False,
                        scaling='density',
                        return_onesided=True,
                        )
                _,Pxy = csd(
                        uL,uR,
                        fs=fs,
                        nperseg=nperseg,
                        noverlap=noverlap,
                        window='hann',
                        detrend=False,
                        scaling='density',
                        return_onesided=True,
                        )
                
                eps = np.finfo(float).eps
                Pxx = np.real(Pxx) ## imag part of auto- cross spectral density is =0 anyway
                Pyy = np.real(Pyy)
                Pxx = np.maximum(Pxx, eps)
                Pyy = np.maximum(Pyy, eps)
                Coh = (np.abs(Pxy)**2) / (Pxx * Pyy)
                return xi,zi,Pxy,Coh
            
            def __coherence_kernel_z(xi,ti,yii,do_density_weighting):
                
                ## 1D [z] □′ or ρ·□″ vectors
                if do_density_weighting:
                    uR = rho[xi,yii,:,ti] * data_R[xi,yii,:,ti]
                else:
                    uR = data_R[xi,yii,:,ti]
                uL = data_L[xi,0,:,ti]
                
                N = uL.shape[0]
                
                _,Pxx = csd(
                        uL,uL,
                        fs=fs,
                        nperseg=N,
                        noverlap=0,
                        window='boxcar',
                        detrend=False,
                        scaling='density',
                        return_onesided=True,
                        )
                _,Pyy = csd(
                        uR,uR,
                        fs=fs,
                        nperseg=N,
                        noverlap=0,
                        window='boxcar',
                        detrend=False,
                        scaling='density',
                        return_onesided=True,
                        )
                _,Pxy = csd(
                        uL,uR,
                        fs=fs,
                        nperseg=N,
                        noverlap=0,
                        window='boxcar',
                        detrend=False,
                        scaling='density',
                        return_onesided=True,
                        )
                
                eps = np.finfo(float).eps
                Pxx = np.real(Pxx) ## imag part of auto- cross spectral density is =0 anyway
                Pyy = np.real(Pyy)
                Pxx = np.maximum(Pxx, eps)
                Pyy = np.maximum(Pyy, eps)
                Coh = (np.abs(Pxy)**2) / (Pxx * Pyy)
                return xi,ti,Pxy,Coh
            
            self.comm.Barrier()
            t_start = timeit.default_timer()
            
            ## [y] loop inner (indices within chunk)
            for yi in range(cy1,cy2):
                
                yii  = yi - cy1 ## chunk local
                yiii = yi - ry1 ## rank local
                
                # ===========================================================================
                # Cross-Correlation [t] : loop over [x,z]
                # ===========================================================================
                
                ## Cross-correlation buffer for [y] loop inner
                R_xz = np.zeros((nx,nz,n_lags_t), dtype=np.float64) ## [x,z] range for ccor(t)
                
                ## Concurrent/threaded execution for ccor(t)
                tasks = [(xi,zi,yii,do_density_weighting) for xi in range(nx) for zi in range(nz)]
                with ThreadPoolExecutor(max_workers=n_threads) as executor:
                    results = executor.map(lambda task: __ccor_kernel_t(*task,), tasks)
                    for xi,zi,result in results:
                        R_xz[xi,zi,:] = result
                
                ## avg in [x,z] & write in rank context
                Rt[scalar][yiii,:] = np.mean(R_xz, axis=(0,1), dtype=np.float64)
                
                # ===========================================================================
                # Coherence [t] : loop over [x,z]
                # ===========================================================================
                
                ## Coherence buffer for [y] loop inner
                Coh_xz = np.zeros((nx,nz,nf), dtype=np.complex128) ## [x,z] range
                P_xz   = np.zeros((nx,nz,nf), dtype=np.complex128) ## [x,z] range
                
                ## Concurrent/threaded execution for Pxy & Coherence
                tasks = [(xi,zi,yii,do_density_weighting) for xi in range(nx) for zi in range(nz)]
                with ThreadPoolExecutor(max_workers=n_threads) as executor:
                    results = executor.map(lambda task: __coherence_kernel_t(*task,), tasks)
                    for xi,zi,P,Coh in results:
                        Coh_xz[xi,zi,:] = Coh[fp]
                        P_xz[xi,zi,:]   = P[fp]
                
                # if np.isnan(Coh_xz).any():
                #     print('NaNs in Coh_xz')
                #     self.comm.Abort(1)
                # if np.isnan(P_xz).any():
                #     print('NaNs in P_xz')
                #     self.comm.Abort(1)
                
                Coh_t[scalar][yiii,:] = np.mean(Coh_xz , axis=(0,1), dtype=np.complex128)
                Pt[scalar][yiii,:]    = np.mean(P_xz   , axis=(0,1), dtype=np.complex128)
                
                # ===========================================================================
                # Coherence [z] : loop over [x,t]
                # ===========================================================================
                
                ## Coherence buffer for [y] loop inner
                Coh_xt = np.zeros((nx,nt,nkz), dtype=np.complex128) ## [x,t] range
                P_xt   = np.zeros((nx,nt,nkz), dtype=np.complex128) ## [x,t] range
                
                ## Concurrent/threaded execution for Pxy & Coherence
                tasks = [(xi,ti,yii,do_density_weighting) for xi in range(nx) for ti in range(nt)]
                with ThreadPoolExecutor(max_workers=n_threads) as executor:
                    results = executor.map(lambda task: __coherence_kernel_z(*task,), tasks)
                    for xi,ti,P,Coh in results:
                        Coh_xt[xi,ti,:] = Coh[kzp]
                        P_xt[xi,ti,:]   = P[kzp]
                
                # if np.isnan(Coh_xt).any():
                #     print('NaNs in Coh_xt')
                #     self.comm.Abort(1)
                # if np.isnan(P_xt).any():
                #     print('NaNs in P_xt')
                #     self.comm.Abort(1)
                
                Coh_z[scalar][yiii,:] = np.mean(Coh_xt , axis=(0,1), dtype=np.complex128)
                Pz[scalar][yiii,:]    = np.mean(P_xt   , axis=(0,1), dtype=np.complex128)
                
                # ===========================================================================
            
            self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            if verbose: tqdm.write(even_print(msg, format_time_string(t_delta), s=True))
            if verbose: progress_bar.update() ## (scalar, [y] chunk) progress
            
            #break ## DEBUG ([y] loop)
        #break ## DEBUG (scalar loop)
    
    if verbose: progress_bar.close()
    self.comm.Barrier()
    if verbose: print(72*'-')
    
    # ==============================================================
    # Write HDF5 (.h5) file
    # ==============================================================
    
    ## Open on rank 0 and write attributes, dimensions, etc.
    if (self.rank==0):
        with h5py.File(fn_h5_out, 'w') as hfw:
            
            ## Write floats,ints as top-level attributes
            for key,val in data.items():
                if isinstance(data[key], (int,np.int32,np.int64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], (float,np.float32,np.float64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], np.ndarray):
                    pass
                else:
                    print(f'key {key} is type {str(type(data[key]))}')
                    self.comm.Abort(1)
            
            ## Write dim arrays
            hfw.create_dataset( 'dims/x'      , data=x      ) ## [m]
            hfw.create_dataset( 'dims/y'      , data=y      ) ## [m]
            hfw.create_dataset( 'dims/z'      , data=z      ) ## [m]
            hfw.create_dataset( 'dims/t'      , data=t      ) ## [s]
            hfw.create_dataset( 'dims/freq'   , data=freq   ) ## [1/s] | [Hz]
            hfw.create_dataset( 'dims/kz'     , data=kz     ) ## [1/m]
            hfw.create_dataset( 'dims/lags_t' , data=lags_t ) ## [s]
            
            ## Initialize datasets
            for scalar in scalars:
                hfw.create_dataset( f'covariance/{scalar}'  , shape=(ny,)         , dtype=np.float64    , chunks=None         , data=np.full((ny,),0.,np.float64)         )
                hfw.create_dataset( f'Rt/{scalar}'          , shape=(ny,n_lags_t) , dtype=np.float64    , chunks=(1,n_lags_t) , data=np.full((ny,n_lags_t),0.,np.float64) )
                hfw.create_dataset( f'Pt/{scalar}'          , shape=(ny,nf)       , dtype=np.complex128 , chunks=(1,nf)       , data=np.full((ny,nf),0.,np.complex128)    )
                hfw.create_dataset( f'Pz/{scalar}'          , shape=(ny,nkz)      , dtype=np.complex128 , chunks=(1,nkz)      , data=np.full((ny,nkz),0.,np.complex128)   )
                hfw.create_dataset( f'Coh_t/{scalar}'       , shape=(ny,nf)       , dtype=np.complex128 , chunks=(1,nf)       , data=np.full((ny,nf),0.,np.complex128)    )
                hfw.create_dataset( f'Coh_z/{scalar}'       , shape=(ny,nkz)      , dtype=np.complex128 , chunks=(1,nkz)      , data=np.full((ny,nkz),0.,np.complex128)   )
            
            ## Initialize datasets 1D [y] mean
            for scalar in avg_Re.dtype.names:
                hfw.create_dataset( f'avg/Re/{scalar}', shape=(ny,), dtype=np.float64, chunks=None, data=np.full((ny,),0.,np.float64) )
            for scalar in avg_Fv.dtype.names:
                hfw.create_dataset( f'avg/Fv/{scalar}', shape=(ny,), dtype=np.float64, chunks=None, data=np.full((ny,),0.,np.float64) )
    
    self.comm.Barrier()
    
    with h5py.File(fn_h5_out, 'a', driver='mpio', comm=self.comm) as hfw:
        
        ## Collectively write covariance,Rt,P,Coh
        for scalar in scalars:
            dset = hfw[f'covariance/{scalar}']
            with dset.collective:
                dset[ry1:ry2] = covariance[scalar][:]
            dset = hfw[f'Rt/{scalar}']
            with dset.collective:
                dset[ry1:ry2,:] = Rt[scalar][:,:]
            dset = hfw[f'Pt/{scalar}']
            with dset.collective:
                dset[ry1:ry2,:] = Pt[scalar][:,:]
            dset = hfw[f'Pz/{scalar}']
            with dset.collective:
                dset[ry1:ry2,:] = Pz[scalar][:,:]
            dset = hfw[f'Coh_t/{scalar}']
            with dset.collective:
                dset[ry1:ry2,:] = Coh_t[scalar][:,:]
            dset = hfw[f'Coh_z/{scalar}']
            with dset.collective:
                dset[ry1:ry2,:] = Coh_z[scalar][:,:]
        
        ## Collectively write 1D [y] avgs (Reynolds,Favre)
        for scalar in avg_Re.dtype.names:
            dset = hfw[f'avg/Re/{scalar}']
            with dset.collective:
                dset[ry1:ry2] = avg_Re[scalar][:]
        for scalar in avg_Fv.dtype.names:
            dset = hfw[f'avg/Fv/{scalar}']
            with dset.collective:
                dset[ry1:ry2] = avg_Fv[scalar][:]
    
    ## Report file contents
    self.comm.Barrier()
    if (self.rank==0):
        even_print( os.path.basename(fn_h5_out) , f'{(os.path.getsize(fn_h5_out)/1024**2):0.1f} [MB]' )
        print(72*'-')
        with h5py.File(fn_h5_out,'r') as hfr:
            h5_print_contents(hfr)
    self.comm.Barrier()
    
    if verbose: print(72*'-')
    if verbose: print('total time : rgd.calc_wall_coh_xpln() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return
